"""Helper functions to extract context from GraphQL error messages."""
import ast
from typing import Optional


def extract_error_context(message: str) -> Optional[dict[str, str]]:
    """Parse the string error message to extract the first context information."""
    try:
        parsed_errors = ast.literal_eval(message)
        return parsed_errors[0].get("extensions").get("context")
    except (SyntaxError, ValueError):
        return None



"""Functions to import assets into an AUDIO project."""

import os
from enum import Enum

from kili.core.helpers import is_url
from kili.domain.project import InputType

from .base import (
    BaseAbstractAssetImporter,
    BatchParams,
    ContentBatchImporter,
)
from .exceptions import ImportValidationError
from .types import AssetLike


class AudioDataType(Enum):
    """Audio data type."""

    LOCAL_FILE = "LOCAL_FILE"
    HOSTED_FILE = "HOSTED_FILE"


class AudioDataImporter(BaseAbstractAssetImporter):
    """Class for importing data into an AUDIO project."""

    @staticmethod
    def get_data_type(assets: list[AssetLike]) -> AudioDataType:
        """Determine the type of data to upload from the service payload."""
        content_array = [asset.get("content", "") for asset in assets]
        has_local_file = any(os.path.exists(content) for content in content_array)  # type: ignore
        has_hosted_file = any(is_url(content) for content in content_array)
        if has_local_file and has_hosted_file:
            raise ImportValidationError(
                """
                Cannot upload hosted data and local files at the same time.
                Please separate the assets into 2 calls
                """
            )
        if has_local_file:
            return AudioDataType.LOCAL_FILE
        return AudioDataType.HOSTED_FILE

    def import_assets(self, assets: list[AssetLike], input_type: InputType):
        """Import AUDIO assets into Kili."""
        self._check_upload_is_allowed(assets)
        data_type = self.get_data_type(assets)
        assets = self.filter_duplicate_external_ids(assets)
        if data_type == AudioDataType.LOCAL_FILE:
            assets = self.filter_local_assets(assets, self.raise_error)
            batch_params = BatchParams(is_hosted=False, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        elif data_type == AudioDataType.HOSTED_FILE:
            batch_params = BatchParams(is_hosted=True, is_asynchronous=False)
            batch_importer = ContentBatchImporter(
                self.kili, self.project_params, batch_params, self.pbar
            )
        else:
            raise ImportValidationError
        return self.import_assets_by_batch(assets, batch_importer)



"""Client presentation methods for cloud storage."""

import logging
from collections.abc import Generator, Iterable
from typing import Literal, Optional, overload

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.cloud_storage import (
    DataConnectionFilters,
    DataConnectionId,
    DataIntegrationFilters,
    DataIntegrationId,
    DataIntegrationPlatform,
    DataIntegrationStatus,
    OrganizationId,
    ProjectId,
)
from kili.domain.types import ListOrTuple
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
    resolve_disable_tqdm,
)
from kili.use_cases.cloud_storage import CloudStorageUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())


@for_all_methods(log_call, exclude=["__init__"])
class CloudStorageClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on cloud storage."""

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "isJsonProcessingEnabled",
            "lastCheck",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[dict, None, None]:
        ...

    @overload
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "isJsonProcessingEnabled",
            "lastCheck",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> list[dict]:
        ...

    @typechecked
    def cloud_storage_connections(
        self,
        cloud_storage_connection_id: Optional[str] = None,
        cloud_storage_integration_id: Optional[str] = None,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "id",
            "isJsonProcessingEnabled",
            "lastCheck",
            "numberOfAssets",
            "selectedFolders",
            "projectId",
        ),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage connections that match a set of criteria.

        Args:
            cloud_storage_connection_id: ID of the cloud storage connection.
            cloud_storage_integration_id: ID of the cloud storage integration.
            project_id: ID of the project.
            fields: All the fields to request among the possible fields for the cloud storage connections.
                See [the documentation](https://api-docs.kili-technology.com/types/objects/data-connection) for all possible fields.
            first: Maximum number of cloud storage connections to return.
            skip: Number of skipped cloud storage connections.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage connections is returned.

        Returns:
            A list or a generator of the cloud storage connections that match the criteria.

        Examples:
            >>> kili.cloud_storage_connections(project_id="789465123")
            [{'id': '123456789', 'isJsonProcessingEnabled': False, 'lastCheck': '2023-02-21T14:49:35.606Z', 'numberOfAssets': 42, 'selectedFolders': ['folder1', 'folder2'], 'projectId': '789465123'}]
        """
        if (
            cloud_storage_connection_id is None
            and cloud_storage_integration_id is None
            and project_id is None
        ):
            raise ValueError(
                "At least one of cloud_storage_connection_id, cloud_storage_integration_id or"
                " project_id must be specified"
            )

        disable_tqdm = resolve_disable_tqdm(disable_tqdm, getattr(self, "disable_tqdm", None))
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        if cloud_storage_connection_id is None:
            data_connections_gen = cloud_storage_use_cases.list_data_connections(
                data_connection_filters=DataConnectionFilters(
                    project_id=ProjectId(project_id) if project_id is not None else None,
                    integration_id=(
                        DataIntegrationId(cloud_storage_integration_id)
                        if cloud_storage_integration_id is not None
                        else None
                    ),
                ),
                fields=fields,
                options=QueryOptions(disable_tqdm, first, skip),
            )
        else:
            data_connections_gen = (
                i
                for i in [
                    cloud_storage_use_cases.get_data_connection(
                        DataConnectionId(cloud_storage_connection_id), fields=fields
                    )
                ]
            )

        if as_generator:
            return data_connections_gen
        return list(data_connections_gen)

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[dict, None, None]:
        ...

    @overload
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> list[dict]:
        ...

    @typechecked
    def cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("name", "id", "platform", "status"),
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.
            fields: All the fields to request among the possible fields for the cloud storage integrations.
                See [the documentation](https://api-docs.kili-technology.com/types/objects/data-integration) for all possible fields.
            first: Maximum number of cloud storage integrations to return.
            skip: Number of skipped cloud storage integrations.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the cloud storage integrations is returned.

        Returns:
            A list or a generator of the cloud storage integrations that match the criteria.

        Examples:
            >>> kili.cloud_storage_integrations()
            [{'name': 'My bucket', 'id': '123456789', 'platform': 'AWS', 'status': 'CONNECTED'}]
        """
        disable_tqdm = resolve_disable_tqdm(disable_tqdm, getattr(self, "disable_tqdm", None))
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        data_integrations_gen = CloudStorageUseCases(self.kili_api_gateway).list_data_integrations(
            data_integration_filters=DataIntegrationFilters(
                status=status,
                id=(
                    DataIntegrationId(cloud_storage_integration_id)
                    if cloud_storage_integration_id is not None
                    else None
                ),
                name=name,
                platform=platform,
                organization_id=(
                    OrganizationId(organization_id) if organization_id is not None else None
                ),
            ),
            fields=fields,
            options=options,
        )

        if as_generator:
            return data_integrations_gen
        return list(data_integrations_gen)

    @typechecked
    def count_cloud_storage_integrations(
        self,
        cloud_storage_integration_id: Optional[str] = None,
        name: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        organization_id: Optional[str] = None,
    ) -> int:
        """Count and return the number of cloud storage integrations that match a set of criteria.

        Args:
            cloud_storage_integration_id: ID of the cloud storage integration.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            organization_id: ID of the organization.

        Returns:
            The number of cloud storage integrations that match the criteria.
        """
        return CloudStorageUseCases(self.kili_api_gateway).count_data_integrations(
            DataIntegrationFilters(
                status=status,
                id=(
                    DataIntegrationId(cloud_storage_integration_id)
                    if cloud_storage_integration_id is not None
                    else None
                ),
                name=name,
                platform=platform,
                organization_id=(
                    OrganizationId(organization_id) if organization_id is not None else None
                ),
            )
        )

    @typechecked
    def add_cloud_storage_connection(
        self,
        project_id: str,
        cloud_storage_integration_id: str,
        selected_folders: Optional[list[str]] = None,
        prefix: Optional[str] = None,
        include: Optional[list[str]] = None,
        exclude: Optional[list[str]] = None,
        is_json_processing_enabled: Optional[bool] = None,
    ) -> dict:
        """Connect a cloud storage to a project. More details about parameters
        can be found in the [documentation](https://docs.kili-technology.com/docs/filtering-assets-from-cloud-storage).

        Args:
            project_id: Id of the project.
            cloud_storage_integration_id: Id of the cloud storage integration.
            selected_folders: List of folders of the data integration to connect to the project.
                If not provided, all folders of the data integration will be connected.
                This option is deprecated and will be removed in the future.
            prefix: Filter files to synchronize based on their base path.
            include: List of pattern used to include files based on their path.
            exclude: List of pattern used to exclude files based on their path.
            is_json_processing_enabled: Whether JSON file synchronization is enabled.
                When True, JSON files are processed during synchronization
                to extract text content or create layered images.

        Returns:
            A dict with the DataConnection Id.
        """
        if selected_folders is not None:
            logger.warning(
                "The selected_folders argument is deprecated and will be removed in the future."
            )

        data_connection_id = CloudStorageUseCases(self.kili_api_gateway).add_data_connection(
            project_id=ProjectId(project_id),
            data_integration_id=DataIntegrationId(cloud_storage_integration_id),
            selected_folders=selected_folders,
            fields=("id",),
            prefix=prefix,
            include=include,
            exclude=exclude,
            is_json_processing_enabled=is_json_processing_enabled,
        )["id"]

        return {"id": data_connection_id}

    @typechecked
    def synchronize_cloud_storage_connection(
        self,
        cloud_storage_connection_id: str,
        delete_extraneous_files: bool = False,
        dry_run: bool = False,
    ) -> dict:
        """Synchronize a cloud storage connection.

        This method will compute differences between the cloud storage connection and the project,
            and then validate the differences.

        If `delete_extraneous_files` is True, it will also delete files that are not in the
            cloud storage integration anymore but that are still in the project.

        Args:
            cloud_storage_connection_id: Id of the cloud storage connection.
            delete_extraneous_files: If True, delete extraneous files.
            dry_run: If True, will not synchronize the data connection but only print the
                differences. This is useful to check the differences before applying them to the
                project.

        Returns:
            A dict with the cloud storage connection Id.
        """
        data_connection_id = DataConnectionId(cloud_storage_connection_id)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        cloud_storage_use_cases.synchronize_data_connection(
            data_connection_id=data_connection_id,
            delete_extraneous_files=delete_extraneous_files,
            dry_run=dry_run,
            logger=logger,
        )

        return cloud_storage_use_cases.get_data_connection(
            data_connection_id=data_connection_id, fields=("numberOfAssets", "projectId")
        )

    @typechecked
    def create_cloud_storage_integration(
        self,
        platform: DataIntegrationPlatform,
        name: str,
        fields: ListOrTuple[str] = (
            "id",
            "name",
            "status",
            "platform",
            "allowedPaths",
        ),
        allowed_paths: Optional[list[str]] = None,
        allowed_projects: Optional[list[str]] = None,
        aws_access_point_arn: Optional[str] = None,
        aws_role_arn: Optional[str] = None,
        aws_role_external_id: Optional[str] = None,
        azure_connection_url: Optional[str] = None,
        azure_is_using_service_credentials: Optional[bool] = None,
        azure_sas_token: Optional[str] = None,
        azure_tenant_id: Optional[str] = None,
        gcp_bucket_name: Optional[str] = None,
        include_root_files: Optional[str] = None,
        internal_processing_authorized: Optional[str] = None,
        s3_access_key: Optional[str] = None,
        s3_bucket_name: Optional[str] = None,
        s3_endpoint: Optional[str] = None,
        s3_region: Optional[str] = None,
        s3_secret_key: Optional[str] = None,
        s3_session_token: Optional[str] = None,
    ) -> dict:
        # pylint: disable=line-too-long
        """Create a cloud storage integration.

        Args:
            fields: All the fields to request among the possible fields for the cloud storage integration.
                See [the documentation](https://api-docs.kili-technology.com/types/objects/data-integration) for all possible fields.
            allowed_paths: List of allowed paths.
            allowed_projects: List of allowed projects.
            aws_access_point_arn: AWS access point ARN.
            aws_role_arn: AWS role ARN.
            aws_role_external_id: AWS role external ID.
            azure_connection_url: Azure connection URL.
            azure_is_using_service_credentials: Whether Azure is using service credentials.
            azure_sas_token: Azure SAS token.
            azure_tenant_id: Azure tenant ID.
            gcp_bucket_name: GCP bucket name.
            include_root_files: Whether to include root files.
            internal_processing_authorized: Whether internal processing is authorized.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            s3_access_key: S3 access key.
            s3_bucket_name: S3 bucket name.
            s3_endpoint: S3 endpoint.
            s3_region: S3 region.
            s3_secret_key: S3 secret key.
            s3_session_token: S3 session token.
        """
        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        return cloud_storage_use_cases.create_data_integration(
            platform=platform,
            name=name,
            fields=fields,
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
        )

    @typechecked
    def update_cloud_storage_integration(
        self,
        cloud_storage_integration_id: str,
        allowed_paths: Optional[list[str]] = None,
        allowed_projects: Optional[list[str]] = None,
        aws_access_point_arn: Optional[str] = None,
        aws_role_arn: Optional[str] = None,
        aws_role_external_id: Optional[str] = None,
        azure_connection_url: Optional[str] = None,
        azure_is_using_service_credentials: Optional[bool] = None,
        azure_sas_token: Optional[str] = None,
        azure_tenant_id: Optional[str] = None,
        gcp_bucket_name: Optional[str] = None,
        include_root_files: Optional[str] = None,
        internal_processing_authorized: Optional[str] = None,
        name: Optional[str] = None,
        organization_id: Optional[str] = None,
        platform: Optional[DataIntegrationPlatform] = None,
        status: Optional[DataIntegrationStatus] = None,
        s3_access_key: Optional[str] = None,
        s3_bucket_name: Optional[str] = None,
        s3_endpoint: Optional[str] = None,
        s3_region: Optional[str] = None,
        s3_secret_key: Optional[str] = None,
        s3_session_token: Optional[str] = None,
    ) -> dict:
        """Update cloud storage data integration.

        Args:
            allowed_paths: List of allowed paths.
            allowed_projects: List of allowed projects.
            aws_access_point_arn: AWS access point ARN.
            aws_role_arn: AWS role ARN.
            aws_role_external_id: AWS role external ID.
            azure_connection_url: Azure connection URL.
            azure_is_using_service_credentials: Whether Azure is using service credentials.
            azure_sas_token: Azure SAS token.
            azure_tenant_id: Azure tenant ID.
            cloud_storage_integration_id: Data integration ID.
            gcp_bucket_name: GCP bucket name.
            include_root_files: Whether to include root files.
            internal_processing_authorized: Whether internal processing is authorized.
            organization_id: Organization ID.
            name: Name of the cloud storage integration.
            platform: Platform of the cloud storage integration.
            status: Status of the cloud storage integration.
            s3_access_key: S3 access key.
            s3_bucket_name: S3 bucket name.
            s3_endpoint: S3 endpoint.
            s3_region: S3 region.
            s3_secret_key: S3 secret key.
            s3_session_token: S3 session token.
        """
        return CloudStorageUseCases(self.kili_api_gateway).update_data_integration(
            data_integration_id=DataIntegrationId(cloud_storage_integration_id),
            name=name,
            platform=platform,
            allowed_paths=allowed_paths,
            allowed_projects=allowed_projects,
            aws_access_point_arn=aws_access_point_arn,
            aws_role_arn=aws_role_arn,
            aws_role_external_id=aws_role_external_id,
            azure_connection_url=azure_connection_url,
            azure_is_using_service_credentials=azure_is_using_service_credentials,
            azure_sas_token=azure_sas_token,
            azure_tenant_id=azure_tenant_id,
            gcp_bucket_name=gcp_bucket_name,
            include_root_files=include_root_files,
            internal_processing_authorized=internal_processing_authorized,
            organization_id=organization_id,
            s3_access_key=s3_access_key,
            s3_bucket_name=s3_bucket_name,
            s3_endpoint=s3_endpoint,
            s3_region=s3_region,
            s3_secret_key=s3_secret_key,
            s3_session_token=s3_session_token,
            status=status,
        )

    @typechecked
    def delete_cloud_storage_integration(self, cloud_storage_integration_id: str) -> str:
        """Delete a cloud storage integration.

        Args:
            cloud_storage_integration_id: Id of the cloud storage integration.
        """
        cloud_storage_integration_id = DataIntegrationId(cloud_storage_integration_id)

        cloud_storage_use_cases = CloudStorageUseCases(self.kili_api_gateway)

        return cloud_storage_use_cases.delete_data_integration(
            data_integration_id=cloud_storage_integration_id
        )



"""Common code for the GeoJson exporter."""

import json
from pathlib import Path

from kili_formats.format.geojson import convert_from_kili_to_geojson_format
from kili_formats.types import Job, JobTool

from kili.domain.ontology import JobMLTask
from kili.services.export.exceptions import NotCompatibleInputType, NotCompatibleOptions
from kili.services.export.format.base import AbstractExporter
from kili.services.export.tools import is_geotiff_asset_with_lat_lon_coords
from kili.utils.tqdm import tqdm


class GeoJsonExporter(AbstractExporter):
    """Common code for GeoJson exporter."""

    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""
        if self.normalized_coordinates is not None:
            raise NotCompatibleOptions(
                "The GeoJson annotation format can only be exported with lat/long coordinates."
                " Please set `normalized_coordinates` to None."
            )

        if self.split_option != "merged":
            raise NotCompatibleOptions(
                "The current implementation only supports merged annotations."
            )

        if self.single_file:
            raise NotCompatibleOptions(
                "The GeoJson annotation format cannot be exported into a single file."
            )

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""
        if self.project["inputType"] not in ["IMAGE", "GEOSPATIAL"]:
            raise NotCompatibleInputType(
                f"Project with input type '{self.project['inputType']}' not compatible with"
                " GeoJson export format."
            )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check if the export label format is compatible with the job."""
        if "tools" not in job:
            return False

        if job["mlTask"] != JobMLTask.OBJECT_DETECTION:
            return False

        compatible_tools = {
            JobTool.RECTANGLE,
            JobTool.POLYGON,
            JobTool.SEMANTIC,
            JobTool.MARKER,
            JobTool.POLYLINE,
        }

        return all(
            tool in compatible_tools
            for tool in job["tools"]  # pyright: ignore[reportGeneralTypeIssues]
        )

    def process_and_save(self, assets: list[dict], output_filename: Path) -> None:
        self.logger.info("Exporting to GeoJson format")

        labels_folder = self.base_folder / "labels"
        labels_folder.mkdir(parents=True, exist_ok=True)
        project_type = self.project.get("inputType")
        geotiff_assets = [
            asset
            for asset in assets
            if project_type == "GEOSPATIAL"
            or is_geotiff_asset_with_lat_lon_coords(asset, self.kili.http_client)
        ]
        if len(geotiff_assets) < len(assets):
            self.logger.warning(
                f"Among {len(assets)} assets, only {len(geotiff_assets)} are geotiff assets and"
                " will be exported."
            )

        # Get json_interface for GIS-friendly property names
        json_interface = self.project.get("jsonInterface")

        for asset in tqdm(geotiff_assets, disable=self.disable_tqdm):
            _process_asset(asset, labels_folder, json_interface, flatten_properties=True)

        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)


def _process_asset(
    asset: dict,
    labels_folder: Path,
    json_interface: dict | None = None,
    flatten_properties: bool = False,
) -> None:
    # Collect all labels to process (handle both latestLabel and latestLabels)
    labels_to_process = []
    if "latestLabel" in asset and asset["latestLabel"]:
        labels_to_process.append(asset["latestLabel"])
    if "latestLabels" in asset and asset["latestLabels"]:
        for label in asset["latestLabels"]:
            if label is not None:
                labels_to_process.append(label)

    if not labels_to_process:
        return

    # Process each label
    for label_idx, latest_label in enumerate(labels_to_process, start=1):
        # Add label suffix if we have multiple labels
        label_suffix = f"_label{label_idx}" if len(labels_to_process) > 1 else ""

        geojson_feature_collection = convert_from_kili_to_geojson_format(
            latest_label["jsonResponse"], json_interface, flatten_properties
        )
        filepath = labels_folder / f"{asset['externalId']}{label_suffix}.geojson"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, "w", encoding="utf-8") as file:
            json.dump(geojson_feature_collection, file)



"""Asset domain."""

from .asset import AssetExternalId, AssetFilters, AssetId, AssetStatus, get_asset_default_fields

__all__ = ["AssetFilters", "AssetId", "AssetExternalId", "AssetStatus", "get_asset_default_fields"]



"""Assets related mutations."""

GQL_APPEND_MANY_ASSETS = """
mutation appendManyAssets(
    $data: AppendManyAssetsData!,
    $where: ProjectWhere!
  ) {
  data: appendManyAssets(
    data: $data,
    where: $where
  ) {
      id
  }
}
"""

GQL_APPEND_MANY_FRAMES_TO_DATASET = """
mutation(
    $data: AppendManyFramesToDatasetAsynchronouslyData!,
    $where: ProjectWhere!
  ) {
  data: appendManyFramesToDatasetAsynchronously(
    data: $data,
    where: $where
  ) {
    id
  }
}
"""

GQL_SET_ASSET_CONSENSUS = """
mutation setAssetConsensus(
    $assetId: ID,
    $externalId: String,
    $projectId: ID!,
    $isConsensus: Boolean!
  ) {
  data: setAssetConsensus(
    assetId: $assetId,
    externalId: $externalId,
    projectId: $projectId,
    isConsensus: $isConsensus
  )
}
"""



"""Copy project implementation."""

import itertools
import logging
import warnings
from typing import TYPE_CHECKING, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.project.types import CopyProjectInput
from kili.domain.asset import AssetFilters
from kili.domain.label import LabelFilters
from kili.domain.project import InputTypeEnum, ProjectId

if TYPE_CHECKING:
    from kili.client import Kili


class ProjectCopier:  # pylint: disable=too-few-public-methods
    """Class for copying an existing project."""

    FIELDS_PROJECT = (
        "title",
        "description",
        "id",
        "dataConnections.id",
        "inputType",
        "jsonInterface",
        "workflowVersion",
    )

    def __init__(self, kili: "Kili") -> None:
        """Simple initialization."""
        self.disable_tqdm = False
        self.kili = kili

    def copy_project(  # pylint: disable=too-many-arguments
        self,
        from_project_id: str,
        title: Optional[str],
        description: Optional[str],
        copy_members: bool,
        copy_assets: bool,
        copy_labels: bool,
        disable_tqdm: Optional[bool],
    ) -> str:
        """Copy an existing project."""
        self.disable_tqdm = disable_tqdm

        logging.basicConfig()
        logger = logging.getLogger("kili.services.copy_project")
        logger.setLevel(logging.INFO)

        if copy_labels:
            if not copy_assets:
                raise ValueError(
                    "`copy_assets` must be set to `True` for copying the source project labels."
                )
            if not copy_members:
                raise ValueError(
                    "`copy_members` must be set to `True` for copying the source project labels."
                )

        fields = self.FIELDS_PROJECT

        src_project = self.kili.kili_api_gateway.get_project(ProjectId(from_project_id), fields)

        if src_project["dataConnections"] and copy_assets:
            raise NotImplementedError("Copying projects with cloud storage is not supported.")

        if src_project["inputType"] in [
            InputTypeEnum.LLM_INSTR_FOLLOWING,
            InputTypeEnum.LLM_RLHF,
            InputTypeEnum.LLM_STATIC,
        ]:
            raise NotImplementedError(
                f"Copying projects with input type {src_project['inputType']} is not supported."
            )

        logger.info("Copying new project...")

        new_project_id = self.kili.kili_api_gateway.copy_project(
            ProjectId(from_project_id),
            CopyProjectInput(
                should_copy_members=copy_members,
                should_copy_assets=copy_assets,
            ),
        )

        logger.info("Created new project %s", new_project_id)

        self.kili.update_properties_in_project(
            project_id=new_project_id,
            title=title or self._generate_project_title(src_project["title"]),
            description=description,
        )

        logger.info("Updated title/description")

        if copy_labels:
            logger.info("Copying labels...")
            if src_project["workflowVersion"] == "V2":
                self._copy_labels(from_project_id=from_project_id, new_project_id=new_project_id)
            else:
                warnings.warn(
                    "Warning: "
                    "copying a project of an old workflow mode may cause problems in asset status and assignation. "
                    "Consider creating a new project instead.",
                    DeprecationWarning,
                )
                self._copy_labels_legacy(
                    from_project_id=from_project_id, new_project_id=new_project_id
                )

        return new_project_id

    def _generate_project_title(self, src_title: str) -> str:
        projects = self.kili.projects(
            fields=["title"], search_query=f"{src_title}%", disable_tqdm=True
        )
        proj_titles = {proj["title"] for proj in projects}
        new_title = f"{src_title} (copy)"
        i = 1
        while new_title in proj_titles:
            new_title = f"{src_title} (copy {i})"
            i += 1
        return new_title

    def _copy_labels(self, from_project_id: str, new_project_id: str) -> None:
        """Method to copy labels from the source project to the new project : applicable for WFV2."""
        nb_labels_to_copy = self.kili.kili_api_gateway.count_labels(
            LabelFilters(project_id=ProjectId(from_project_id))
        )

        if nb_labels_to_copy == 0:
            return

        assets_src_project = self.kili.kili_api_gateway.list_assets(
            AssetFilters(project_id=ProjectId(from_project_id)),
            ["id", "externalId", "labels.id"],
            QueryOptions(disable_tqdm=True),
        )

        assets_dst_project = self.kili.kili_api_gateway.list_assets(
            AssetFilters(project_id=ProjectId(new_project_id)),
            ["id", "externalId"],
            QueryOptions(disable_tqdm=True),
        )

        # Iterate on assets of the source project
        # to copy labels to the new projectq
        assets_src_project_list = [
            asset
            for asset in assets_src_project
            if "labels" in asset and asset["labels"] and len(asset["labels"]) > 0
        ]
        assets_dst_project_map = {asset["externalId"]: asset["id"] for asset in assets_dst_project}

        for i, src_asset in enumerate(assets_src_project_list, start=1):
            is_last = i == len(assets_src_project_list)
            skip_rebuilds = not is_last
            src_asset_id = src_asset["id"]
            dst_asset_id = assets_dst_project_map.get(src_asset["externalId"])
            if not dst_asset_id:
                raise ValueError(
                    f"Asset with externalId {src_asset['externalId']} not found in new project {new_project_id}."
                )

            self.kili.kili_api_gateway.copy_labels(
                src_asset_id=src_asset_id,
                dst_asset_id=dst_asset_id,
                project_id=new_project_id,
                skip_rebuilds=skip_rebuilds,
            )

    # pylint: disable=too-many-locals
    def _copy_labels_legacy(self, from_project_id: str, new_project_id: str) -> None:
        """Legacy mlethod to copy labels from the source project to the new project : applicable for WFV1.

        !!! warning
            This method is deprecated and will be removed in the next major release.
            Asset with send back labels are not supported (status wise) and asset assignation is not supported.
        """
        assets_new_project = self.kili.kili_api_gateway.list_assets(
            AssetFilters(project_id=ProjectId(new_project_id)),
            ["id", "externalId"],
            QueryOptions(disable_tqdm=True),
        )
        assets_new_project_map = {asset["externalId"]: asset["id"] for asset in assets_new_project}

        members_new_project = self.kili.project_users(
            project_id=new_project_id, fields=["user.email", "user.id"], disable_tqdm=True
        )
        members_new_project_map = {
            member["user"]["email"]: member["user"]["id"] for member in members_new_project
        }

        nb_labels_to_copy = self.kili.kili_api_gateway.count_labels(
            LabelFilters(project_id=ProjectId(from_project_id))
        )
        if nb_labels_to_copy == 0:
            return

        labels = self.kili.labels(
            project_id=from_project_id,
            fields=[
                "author.email",
                "jsonResponse",
                "jsonResponseUrl",
                "secondsToLabel",
                "isLatestLabelForUser",
                "labelOf.externalId",
                "labelType",
                "modelName",
            ],
            disable_tqdm=True,
        )
        labels = [
            label for label in labels if (label["isLatestLabelForUser"] or label.get("modelName"))
        ]

        # `append_labels` does not take arrays for `model_name` and `label_type` arguments
        # we need to sort and group the labels by `model_name` and `label_type`
        # and upload the grouped labels by batch to `append_labels`
        labels = sorted(
            labels,
            key=lambda label: (
                label["labelType"],
                label["modelName"] is None,
                label["isLatestLabelForUser"],
                label["modelName"],
            ),
        )
        labels_iterator = itertools.groupby(
            labels,
            key=lambda label: (
                label["labelType"],
                label["modelName"] is None,
                label["isLatestLabelForUser"],
                label["modelName"],
            ),
        )

        for key, group in labels_iterator:
            label_type, _, _, model_name = key
            group = list(group)

            # map external id of source project asset to
            # internal id of new project corresponding asset
            # since both source_project and new_project have the same externalIds
            asset_id_array = [
                assets_new_project_map[label["labelOf"]["externalId"]] for label in group
            ]
            json_response_array = [label["jsonResponse"] for label in group]
            author_id_array = [members_new_project_map[label["author"]["email"]] for label in group]
            seconds_to_label_array = [label["secondsToLabel"] for label in group]

            self.kili.append_labels(
                asset_id_array=asset_id_array,
                json_response_array=json_response_array,
                author_id_array=author_id_array,
                seconds_to_label_array=seconds_to_label_array,
                model_name=model_name,
                label_type=label_type,
                disable_tqdm=False,
            )



"""GraphQL Project Workflow operations."""


def get_update_project_workflow_mutation(fragment: str) -> str:
    """Return the GraphQL editProjectWorkflowSettings mutation."""
    return f"""
        mutation editProjectWorkflowSettings($input: EditProjectWorkflowSettingsInput!) {{
            data: editProjectWorkflowSettings(input: $input) {{
                {fragment}
            }}
        }}
        """


def get_steps_query(fragment: str) -> str:
    """Return the GraphQL getSteps query."""
    return f"""
        query getSteps($where: ProjectWhere!, $first: PageSize!, $skip: Int!) {{
            data: projects(where: $where, first: $first, skip: $skip) {{
            {fragment}
            }}
        }}
        """



"""Queries of asset mutations."""

from kili.entrypoints.mutations.project.fragments import PROJECT_FRAGMENT_ID

GQL_ASSIGN_ASSETS = """
mutation assignAssetsToLabelers(
    $where: AssetWhere!,
    $userIds: [String!]!
) {
    data: assignAssetsToLabelers(
        where: $where,
        userIds: $userIds
    ) {
      id
    }
  }
"""

GQL_UPDATE_PROPERTIES_IN_ASSETS = """
mutation(
    $whereArray: [AssetWhere!]!
    $dataArray: [AssetData!]!
) {
  data: updatePropertiesInAssets(
    where: $whereArray,
    data: $dataArray
  ) {
    id
  }
}
"""

GQL_DELETE_MANY_FROM_DATASET = f"""
mutation($where: AssetWhere!) {{
  data: deleteManyFromDataset(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_ADD_ALL_LABELED_ASSETS_TO_REVIEW = f"""
mutation($where: AssetWhere!) {{
  data: addAllLabeledAssetsToReview(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_SEND_BACK_ASSETS_TO_QUEUE = f"""
mutation($where: AssetWhere!) {{
  data: sendBackAssetsToQueue(where: $where) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
"""

GQL_SKIP_ASSET = """
mutation SkipAsset($reason: String!, $where: AssetWhere!) {
  skipAsset(reason: $reason, where: $where) {
    id
  }
}
"""

GQL_UNSKIP_ASSET = """
mutation UnskipAsset($projectId: ID!, $assetId: ID!) {
  unskipAsset(projectId: $projectId, assetId: $assetId)
}
"""



from typing import Optional, cast

from kili_formats import convert_from_kili_to_llm_static_or_dynamic_format
from kili_formats.types import ChatItem, Conversation, JobLevel

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway

CHAT_ITEMS_NEEDED_FIELDS = [
    "id",
    "content",
    "modelId",
    "parentId",
    "role",
]

LABELS_NEEDED_FIELDS = [
    "annotations.id",
    "author.id",
    "author.email",
    "author.firstname",
    "author.lastname",
    *(f"chatItems.{field}" for field in CHAT_ITEMS_NEEDED_FIELDS),
    "createdAt",
    "id",
    "isLatestLabelForUser",
    "jsonResponse",
    "labelType",
    "modelName",
]


DEFAULT_JOB_LEVEL = JobLevel.ROUND


def get_model_name(model_id: Optional[str], project_models: list[dict]) -> Optional[str]:
    try:
        return next(
            model["configuration"]["model"] for model in project_models if model["id"] == model_id
        )
    except (KeyError, StopIteration):
        return model_id


class LLMExporter:
    """Handle exports of LLM_STATIC and LLM_INSTR_FOLLOWING projects."""

    def __init__(self, kili_api_gateway: KiliAPIGateway):
        self.kili_api_gateway = kili_api_gateway

    def export(self, assets: list[dict], json_interface: dict) -> list[Conversation]:
        return [self.format_asset(asset, json_interface) for asset in assets if asset.get("labels")]

    def format_asset(self, asset: dict, json_interface: dict) -> Conversation:
        label = asset["labels"][-1]
        chat_items = [
            {
                "id": chat_item["id"],
                "content": chat_item.get("content"),
                "externalId": chat_item.get("externalId") or chat_item["id"],
                "modelName": chat_item.get("modelName")
                or get_model_name(chat_item.get("modelId"), asset["assetProjectModels"]),
                "role": chat_item.get("role"),
            }
            for chat_item in (
                label.get("chatItems") or self.kili_api_gateway.list_chat_items(asset["id"])
            )
        ]

        metadata = {}
        if asset.get("assetProjectModels"):
            metadata["models"] = asset["assetProjectModels"]

        chat_items_without_ids = [
            cast(ChatItem, {k: v for k, v in chat_item.items() if k != "id"})
            for chat_item in chat_items
        ]

        return {
            "chatItems": chat_items_without_ids,
            "externalId": asset["externalId"],
            "label": convert_from_kili_to_llm_static_or_dynamic_format(
                annotations=label["annotations"],
                chat_items=chat_items,
                jobs=json_interface["jobs"],
            ),
            "labeler": label["author"]["email"],
            "metadata": metadata,
        }



"""DEPRECATED.

From https://docs.kili-technology.com/reference/graphql-api#enums.
"""

from typing import Literal

LicenseType = Literal[
    "DISCOVERY",
    "FREEMIUM",
    "PAID",
]


LockType = Literal[
    "REVIEW",
    "DEFAULT",
]


NotificationStatus = Literal[
    "FAILURE",
    "PENDING",
    "SUCCESS",
]


OrganizationRole = Literal[
    "ADMIN",
    "USER",
]


ProjectRole = Literal[
    "ADMIN",
    "TEAM_MANAGER",
    "REVIEWER",
    "LABELER",
]


DemoProjectType = Literal[
    "DEMO_COMPUTER_VISION_TUTORIAL",
    "DEMO_TEXT_TUTORIAL",
    "DEMO_PDF_TUTORIAL",
    "VIDEO_FRAME_OBJECT_TRACKING",
    "DEMO_SEGMENTATION_COCO",
    "DEMO_NER",
    "DEMO_ID_OCR",
    "DEMO_REVIEWS",
    "DEMO_OCR",
    "DEMO_NER_TWEETS",
    "DEMO_PLASTIC",
    "DEMO_CHATBOT",
    "DEMO_PDF",
    "DEMO_ANIMALS",
    "DEMO_LLM",
    "DEMO_LLM_INSTR_FOLLOWING",
    "DEMO_SEGMENTATION",
]



"""Types for the ProjectWorkflow-related Kili API gateway functions."""

from dataclasses import dataclass
from typing import Optional

from kili.domain.project import WorkflowStepCreate, WorkflowStepUpdate


@dataclass
class ProjectWorkflowDataKiliAPIGatewayInput:
    """ProjectWorkflow input data for Kili API Gateway."""

    enforce_step_separation: Optional[bool]
    create_steps: Optional[list[WorkflowStepCreate]]
    update_steps: Optional[list[WorkflowStepUpdate]]
    delete_steps: Optional[list[str]]



"""GraphQL payload data mappers for project operations."""


from kili.domain.project import ProjectFilters

from .types import ProjectDataKiliAPIGatewayInput


def project_where_mapper(filters: ProjectFilters) -> dict:
    """Build the GraphQL ProjectWhere variable to be sent in an operation."""
    ret = {
        "id": filters.id,
        "searchQuery": filters.search_query,
        "shouldRelaunchKpiComputation": filters.should_relaunch_kpi_computation,
        "starred": filters.starred,
        "updatedAtGte": filters.updated_at_gte,
        "updatedAtLte": filters.updated_at_lte,
        "createdAtGte": filters.created_at_gte,
        "createdAtLte": filters.created_at_lte,
        "organizationId": filters.organization_id,
        "tagIds": filters.tag_ids,
        "deleted": filters.deleted,
    }
    if filters.archived is not None:
        ret["archived"] = filters.archived
    return ret


def project_data_mapper(data: ProjectDataKiliAPIGatewayInput) -> dict:
    """Build the GraphQL ProjectData variable to be sent in an operation."""
    result = {
        "archived": data.archived,
        "author": data.author,
        "complianceTags": data.compliance_tags,
        "consensusMark": data.consensus_mark,
        "consensusTotCoverage": data.consensus_tot_coverage,
        "description": data.description,
        "canNavigateBetweenAssets": data.can_navigate_between_assets,
        "canSkipAsset": data.can_skip_asset,
        "honeypotMark": data.honeypot_mark,
        "instructions": data.instructions,
        "jsonInterface": data.json_interface,
        "minConsensusSize": data.min_consensus_size,
        "reviewCoverage": data.review_coverage,
        "shouldAutoAssign": data.should_auto_assign,
        "shouldRelaunchKpiComputation": data.should_relaunch_kpi_computation,
        "title": data.title,
        "useHoneyPot": data.use_honeypot,
    }

    if data.metadata_properties is not None:
        result["metadataProperties"] = data.metadata_properties
    elif data.metadata_types is not None:
        metadata_properties = {}
        for key, type_value in data.metadata_types.items():
            metadata_properties[key] = {
                "filterable": True,
                "type": type_value,
                "visibleByLabeler": True,
                "visibleByReviewer": True,
            }
        result["metadataProperties"] = metadata_properties

    return result



"""Configuration file loader for the Kili SDK."""

import json
import warnings
from pathlib import Path
from typing import Any, Optional


def load_config_from_file(
    filename: str = "kili-sdk-config.json",
    search_paths: Optional[list[Path]] = None,
) -> dict[str, Any]:
    """Load configuration from a JSON file.

    Searches for the configuration file in the provided search paths.
    If no search paths are provided, searches in the current working directory
    and the user's home directory.

    Args:
        filename: Name of the configuration file. Defaults to "kili-sdk-config.json".
        search_paths: List of directories to search for the configuration file.
            If None, defaults to [cwd, home].

    Returns:
        A dictionary containing the configuration, or an empty dict if no file is found.
    """
    if search_paths is None:
        search_paths = [
            Path.cwd(),
            Path.home(),
        ]

    for search_path in search_paths:
        config_path = search_path / filename
        if config_path.exists():
            try:
                with open(config_path, encoding="utf-8") as f:
                    return json.load(f)
            except json.JSONDecodeError as e:
                warnings.warn(
                    f"Invalid JSON in configuration file '{config_path}': {e}. Skipping this file",
                    UserWarning,
                    stacklevel=2,
                )
                continue
            except OSError as e:
                warnings.warn(
                    f"Could not read configuration file '{config_path}': {e}. Skipping this file.",
                    UserWarning,
                    stacklevel=2,
                )
                continue

    return {}



"""Module for common argument validators across client methods."""

import warnings
from typing import Any, Optional

from kili.domain.types import ListOrTuple


def disable_tqdm_if_as_generator(
    as_generator: bool, disable_tqdm: Optional[bool]
) -> Optional[bool]:
    """Disable tqdm in user-facing queries method if the return type is asked as a generator."""
    if as_generator and not disable_tqdm:
        disable_tqdm = True
        warnings.warn(
            "tqdm has been forced disabled because its behavior is not compatible with the"
            " generator return type",
            stacklevel=2,
        )
    return disable_tqdm


def resolve_disable_tqdm(
    disable_tqdm: bool | None, client_disable_tqdm: bool | None
) -> bool | None:
    """Resolve the disable_tqdm parameter with priority: function param > client global setting.

    Args:
        disable_tqdm: The disable_tqdm parameter passed to the function.
        client_disable_tqdm: The global disable_tqdm setting from the client.
            Can be None if the client doesn't have this attribute set (e.g., in tests).

    Returns:
        The resolved disable_tqdm value, or None if both are None (to use function default).

    Examples:
        >>> resolve_disable_tqdm(True, False)  # Function param takes priority
        True
        >>> resolve_disable_tqdm(None, True)   # Use client global setting
        True
        >>> resolve_disable_tqdm(None, None)   # Use function default
        None
    """
    if disable_tqdm is not None:
        return disable_tqdm
    return client_disable_tqdm


def assert_all_arrays_have_same_size(
    arrays: list[Optional[ListOrTuple[Any]]], raise_error: bool = True
) -> bool:
    """Assert that all given arrays have the same size if they are not None."""
    sizes_arrays = {len(array) for array in arrays if array is not None}
    if len(sizes_arrays) > 1:
        if raise_error:
            raise ValueError("All arrays should have the same length")
        return False
    return True



"""GraphQL Event operations."""


def get_events_query(fragment: str) -> str:
    """Return the GraphQL events query."""
    return f"""
          query Events($where: EventsWhere!, $pagination: EventPagination!, $order: Order) {{
            data: events(where: $where, pagination: $pagination, order: $order) {{
                {fragment}
              }}
            }}
        """



"""Assets domain namespace for the Kili Python SDK."""
# pylint: disable=too-many-lines,too-many-public-methods

import warnings
from collections.abc import Generator
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Literal,
    Optional,
    TypedDict,
    Union,
    cast,
    overload,
)

from typeguard import typechecked
from typing_extensions import deprecated

from kili.core.helpers import is_url
from kili.domain.asset import (
    AssetStatus,
)
from kili.domain.asset.asset import StatusInStep
from kili.domain.issue import IssueStatus, IssueType
from kili.domain.label import LabelType
from kili.domain.types import ListOrTuple
from kili.domain_api.base import DomainNamespace
from kili.domain_api.namespace_utils import get_available_methods

if TYPE_CHECKING:
    import pandas as pd


class AssetFilter(TypedDict, total=False):
    """Filter options for querying assets.

    This TypedDict defines all available filter parameters that can be used
    when listing or counting assets. All fields are optional.

    Use this filter with `kili.assets.list()` and `kili.assets.count()` methods
    to filter assets based on various criteria such as status, assignee, labels,
    metadata, and more.
    """

    asset_id_in: Optional[list[str]]
    asset_id_not_in: Optional[list[str]]
    assignee_in: Optional[ListOrTuple[str]]
    assignee_not_in: Optional[ListOrTuple[str]]
    consensus_mark_gt: Optional[float]
    consensus_mark_gte: Optional[float]
    consensus_mark_lt: Optional[float]
    consensus_mark_lte: Optional[float]
    created_at_gte: Optional[str]
    created_at_lte: Optional[str]
    external_id_in: Optional[list[str]]
    external_id_strictly_in: Optional[list[str]]
    honeypot_mark_gt: Optional[float]
    honeypot_mark_gte: Optional[float]
    honeypot_mark_lt: Optional[float]
    honeypot_mark_lte: Optional[float]
    inference_mark_gte: Optional[float]
    inference_mark_lte: Optional[float]
    issue_status: Optional[IssueStatus]
    issue_type: Optional[IssueType]
    label_author_in: Optional[list[str]]
    label_category_search: Optional[str]
    label_consensus_mark_gt: Optional[float]
    label_consensus_mark_gte: Optional[float]
    label_consensus_mark_lt: Optional[float]
    label_consensus_mark_lte: Optional[float]
    label_created_at_gt: Optional[str]
    label_created_at_gte: Optional[str]
    label_created_at_lt: Optional[str]
    label_created_at_lte: Optional[str]
    label_created_at: Optional[str]
    label_honeypot_mark_gt: Optional[float]
    label_honeypot_mark_gte: Optional[float]
    label_honeypot_mark_lt: Optional[float]
    label_honeypot_mark_lte: Optional[float]
    label_labeler_in: Optional[ListOrTuple[str]]
    label_labeler_not_in: Optional[ListOrTuple[str]]
    label_reviewer_in: Optional[ListOrTuple[str]]
    label_reviewer_not_in: Optional[ListOrTuple[str]]
    label_type_in: Optional[list[LabelType]]
    metadata_where: Optional[dict[str, Any]]
    skipped: Optional[bool]
    status_in: Optional[list[AssetStatus]]
    step_name_in: Optional[list[str]]
    step_name_not_in: Optional[list[str]]
    step_status_in: Optional[list[StatusInStep]]
    step_status_not_in: Optional[list[StatusInStep]]
    updated_at_gte: Optional[str]
    updated_at_lte: Optional[str]


class VideoProcessingParameters(TypedDict, total=False):
    """Processing parameters for video assets.

    These parameters control how video assets are processed and displayed in Kili.

    Attributes:
        frames_played_per_second: Frame rate for video playback (frames per second)
        number_of_frames: Total number of frames in the video
        start_time: Starting time offset in seconds
    """

    frames_played_per_second: int

    number_of_frames: int

    start_time: float


class ImageLayerParam(TypedDict, total=False):
    """Parameters of an image or geospatial layer.

    Attributes:
        path: local path to the file to use as a layer
        name: Optional name for the layer
    """

    path: str
    name: Optional[str]


class GeospatialLayerParam(ImageLayerParam, total=False):
    """Parameters for geospatial layers, extending ImageLayerParam.

    Attributes:
        path: URL or local path to the geospatial layer file
        name: Optional name for the layer
        bounds: Optional bounding box coordinates [[minX, minY], [maxX, maxY]]
        epsg: Optional coordinate reference system (EPSG3857 or EPSG4326)
    """

    bounds: Optional[list[list[float]]]
    epsg: Optional[Literal["EPSG3857", "EPSG4326"]]


def convert_to_web_layer(layer: GeospatialLayerParam) -> dict[str, Any]:
    """Convert a GeospatialLayerParam to web layer format for URL-based geospatial layers.

    Args:
        layer: Layer parameter dictionary with name, path, bounds, and epsg

    Returns:
        Dictionary with web layer configuration including tileLayerUrl and coordinates settings
    """
    res = {
        "bounds": layer.get("bounds"),
        "epsg": layer.get("epsg") if layer.get("epsg") else "EPSG3857",
        "name": layer.get("name"),
        "tileLayerUrl": layer.get("path"),
        "useClassicCoordinates": False,
        "isBaseLayer": False,
    }
    return res


def convert_to_local_layer(layer: GeospatialLayerParam) -> dict[str, Any]:
    """Convert a GeospatialLayerParam to local layer format for file-based geospatial layers.

    Args:
        layer: Layer parameter dictionary with name and path

    Returns:
        Dictionary with local layer configuration containing name and path
    """
    if layer.get("bounds"):
        warnings.warn(
            "Custom bounds are not yet supported for local layers.",
            stacklevel=1,
        )
    if layer.get("epsg"):
        warnings.warn(
            "Forced epsg are not yet supported for local layers.",
            stacklevel=1,
        )
    res = {
        "name": layer.get("name"),
        "path": layer.get("path"),
    }
    return res


def _snake_to_camel_case(snake_str: str) -> str:
    """Convert snake_case string to camelCase.

    Args:
        snake_str: String in snake_case format

    Returns:
        String in camelCase format
    """
    components = snake_str.split("_")
    return components[0] + "".join(x.title() for x in components[1:])


def _transform_processing_parameters(
    params: dict[str, Any],
) -> dict[str, Any]:
    """Transform processing parameter keys from snake_case to camelCase.

    Args:
        params: Processing parameters with snake_case keys (video or GeoTIFF)

    Returns:
        Dictionary with camelCase keys
    """
    return {_snake_to_camel_case(key): value for key, value in params.items()}


def _prepare_video_processing_parameters(
    params: VideoProcessingParameters, use_native_video: bool
) -> dict[str, Any]:
    """Prepare video processing parameters with defaults.

    Transforms keys from snake_case to camelCase and adds default parameters:
    - shouldUseNativeVideo: True for native video, False for frame-based video
    - shouldKeepNativeFrameRate: False (if framesPlayedPerSecond is specified)

    Args:
        params: Video processing parameters with snake_case keys
        use_native_video: True for native video, False for frame-based video

    Returns:
        Dictionary with camelCase keys and default parameters added
    """
    # Transform to camelCase
    transformed = _transform_processing_parameters(cast(dict[str, Any], params))

    # Add shouldUseNativeVideo based on the method
    transformed["shouldUseNativeVideo"] = use_native_video

    # Add shouldKeepNativeFrameRate=False if framesPlayedPerSecond is defined
    if "framesPlayedPerSecond" in transformed:
        transformed["shouldKeepNativeFrameRate"] = False

    return transformed


class AssetsNamespace(DomainNamespace):  # pylint: disable=too-many-public-methods
    """Assets domain namespace providing asset-related operations.

    This namespace provides access to all asset-related functionality
    including creating, updating, querying, and managing assets.

    The namespace provides the following main operations:
    - list(): Query and list assets
    - count(): Count assets matching filters
    - create_image(): Create image assets
    - create_video_native(): Create video assets from video files
    - create_video_frame(): Create video assets from frame sequences
    - create_geospatial(): Create multi-layer geospatial imagery assets
    - create_pdf(): Create PDF assets
    - create_text(): Create plain text assets
    - create_rich_text(): Create rich-text formatted text assets
    - create_audio(): Create audio assets
    - delete(): Delete assets from projects
    - add_metadata(): Add metadata to assets
    - set_metadata(): Set metadata on assets
    - update_external_id(): Update asset external IDs
    - update_processing_parameter(): Update video processing parameters
    - invalidate(): Send assets back to queue (invalidate current step)
    - move_to_next_step(): Move assets to the next workflow step
    - assign(): Assign assets to labelers
    - update_priority(): Update asset priorities
    - skip(): Skip an asset
    - unskip(): Unskip an asset
    - add_consensus(): Activate or deactivate consensus on an asset

    Examples:
        >>> kili = Kili()
        >>> # List assets
        >>> assets = kili.assets.list(project_id="my_project")

        >>> # Count assets
        >>> count = kili.assets.count(project_id="my_project")

        >>> # Create image assets
        >>> result = kili.assets.create_image(
        ...     project_id="my_project",
        ...     content_array=["https://example.com/image.png"]
        ... )

        >>> # Create video from video file
        >>> result = kili.assets.create_video_native(
        ...     project_id="my_project",
        ...     content="https://example.com/video.mp4",
        ...     processing_parameters={"frames_played_per_second": 25}
        ... )

        >>> # Add asset metadata
        >>> kili.assets.add_metadata(
        ...     json_metadata={"key": "value"},
        ...     project_id="my_project",
        ...     asset_id="asset_id"
        ... )

        >>> # Assign assets to labelers
        >>> kili.assets.assign(
        ...     asset_ids=["asset_id"],
        ...     to_be_labeled_by_array=[["user_id"]]
        ... )
    """

    def __init__(self, client, gateway):
        """Initialize the assets namespace.

        Args:
            client: The Kili client instance
            gateway: The KiliAPIGateway instance for API operations
        """
        super().__init__(client, gateway, "assets")

    @deprecated(
        "'assets' is a namespace, not a callable method. "
        "Use kili.assets.list() or other available methods instead."
    )
    def __call__(self, *args, **kwargs):
        """Raise a helpful error when namespace is called like a method.

        This provides guidance to users migrating from the legacy API
        where assets were accessed via kili.assets(...) to the new domain API
        where they use kili.assets.list(...) or other methods.
        """
        available_methods = get_available_methods(self)
        methods_str = ", ".join(f"kili.{self._domain_name}.{m}()" for m in available_methods)
        raise TypeError(
            f"'{self._domain_name}' is a namespace, not a callable method. "
            f"The legacy API 'kili.{self._domain_name}(...)' has been replaced with the domain API.\n"
            f"Available methods: {methods_str}\n"
            f"Example: kili.{self._domain_name}.list(...)"
        )

    @typechecked
    def list(
        self,
        project_id: str,
        disable_tqdm: Optional[bool] = None,
        download_media: bool = False,
        fields: Optional[ListOrTuple[str]] = None,
        filter: Optional[AssetFilter] = None,
        first: Optional[int] = None,
        format: Optional[str] = None,
        label_output_format: Literal["dict", "parsed_label"] = "dict",
        local_media_dir: Optional[str] = None,
        skip: int = 0,
    ) -> Union[list[dict], "pd.DataFrame"]:
        """List assets from a project.

        Args:
            project_id: Identifier of the project.
            skip: Number of assets to skip (ordered by creation date).
            fields: List of fields to return. If None, returns default fields.
            filter: Additional asset filters to apply (see `AssetFilter` for available keys).
            disable_tqdm: If True, the progress bar will be disabled.
            first: Maximum number of assets to return.
            format: Output format; when set to `"pandas"` returns a DataFrame.
            download_media: If True, downloads media files locally.
            local_media_dir: Directory used when `download_media` is True.
            label_output_format: Format of the returned labels ("dict" or "parsed_label").

        Returns:
            A list of assets or a pandas DataFrame depending on `format`.
        """
        filter_kwargs = filter or {}
        return self._client.assets(
            as_generator=False,
            disable_tqdm=disable_tqdm,
            download_media=download_media,
            fields=fields,
            first=first,
            format=format,
            label_output_format=label_output_format,
            local_media_dir=local_media_dir,
            project_id=project_id,
            skip=skip,
            **filter_kwargs,
        )

    @typechecked
    def list_as_generator(
        self,
        project_id: str,
        disable_tqdm: Optional[bool] = None,
        download_media: bool = False,
        fields: Optional[ListOrTuple[str]] = None,
        filter: Optional[AssetFilter] = None,
        first: Optional[int] = None,
        label_output_format: Literal["dict", "parsed_label"] = "dict",
        local_media_dir: Optional[str] = None,
        skip: int = 0,
    ) -> Generator[dict, None, None]:
        """List assets from a project.

        Args:
            project_id: Identifier of the project.
            skip: Number of assets to skip (ordered by creation date).
            fields: List of fields to return. If None, returns default fields.
            filter: Additional asset filters to apply (see `AssetFilter` for available keys).
            disable_tqdm: If True, the progress bar will be disabled.
            first: Maximum number of assets to return.
            download_media: If True, downloads media files locally.
            local_media_dir: Directory used when `download_media` is True.
            label_output_format: Format of the returned labels ("dict" or "parsed_label").

        Returns:
            A generator of a list of assets.
        """
        filter_kwargs = filter or {}
        return self._client.assets(
            as_generator=True,
            disable_tqdm=disable_tqdm,
            download_media=download_media,
            fields=fields,
            first=first,
            label_output_format=label_output_format,
            local_media_dir=local_media_dir,
            project_id=project_id,
            skip=skip,
            **filter_kwargs,
        )

    @typechecked
    def count(
        self,
        project_id: str,
        filter: Optional[AssetFilter] = None,
    ) -> int:
        """Count assets in a project.

        Args:
            project_id: Identifier of the project.
            filter: Additional asset filters to apply (see `AssetFilter` for available keys).

        Returns:
            The number of assets matching the filters.

        Examples:
            >>> # Count all assets in project
            >>> count = kili.assets.count(project_id="my_project")

            >>> # Count assets with specific status
            >>> count = kili.assets.count(
            ...     project_id="my_project",
            ...     filter={"status_in": ["TODO", "ONGOING"]}
            ... )
        """
        filter_kwargs = filter or {}
        return self._client.count_assets(
            project_id=project_id,
            **filter_kwargs,
        )

    @overload
    def create_image(
        self,
        *,
        project_id: str,
        content: Union[str, dict],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_image(
        self,
        *,
        project_id: str,
        content_array: Union[List[str], List[dict]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_image(
        self,
        *,
        project_id: str,
        content: Optional[Union[str, dict]] = None,
        content_array: Optional[Union[List[str], List[dict]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create image assets in a project.

        Args:
            project_id: Identifier of the project
            content: URL or local file path to an image
            content_array: List of URLs or local file paths to images
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single image asset
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content="https://example.com/image.png"
            ... )

            >>> # Create multiple image assets
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content_array=["https://example.com/image1.png", "https://example.com/image2.png"]
            ... )

            >>> # Create single asset with metadata
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content="https://example.com/image.png",
            ...     json_metadata={"description": "Sample image"}
            ... )
        """
        # Convert singular to plural
        if content is not None:
            content_array = cast(Union[list[str], list[dict]], [content])
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            content_array=content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_layered_image(
        self,
        *,
        project_id: str,
        layers: List[ImageLayerParam],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_layered_image(
        self,
        *,
        project_id: str,
        layers_array: List[List[ImageLayerParam]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_layered_image(
        self,
        *,
        project_id: str,
        layers: Optional[List[ImageLayerParam]] = None,
        layers_array: Optional[List[List[ImageLayerParam]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create image assets in a project.

        Args:
            project_id: Identifier of the project
            layers: List of image layers for a single asset
            layers_array: List of image layer lists for multiple assets
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single image asset
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content="https://example.com/image.png"
            ... )

            >>> # Create multiple image assets
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content_array=["https://example.com/image1.png", "https://example.com/image2.png"]
            ... )

            >>> # Create single asset with metadata
            >>> result = kili.assets.create_image(
            ...     project_id="my_project",
            ...     content="https://example.com/image.png",
            ...     json_metadata={"description": "Sample image"}
            ... )
        """
        # Convert singular to plural
        if layers is not None:
            layers_array = [layers]
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            multi_layer_content_array=cast(Optional[list[list[dict]]], layers_array),
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_video_native(
        self,
        *,
        project_id: str,
        content: Union[str, dict],
        processing_parameters: Optional[VideoProcessingParameters] = None,
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_video_native(
        self,
        *,
        project_id: str,
        content_array: Union[List[str], List[dict]],
        processing_parameters_array: Optional[List[VideoProcessingParameters]] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_video_native(
        self,
        *,
        project_id: str,
        content: Optional[Union[str, dict]] = None,
        content_array: Optional[Union[List[str], List[dict]]] = None,
        processing_parameters: Optional[VideoProcessingParameters] = None,
        processing_parameters_array: Optional[List[VideoProcessingParameters]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create video assets from video files in a project.

        If processing parameters are incomplete, Kili will probe the videos to determine missing parameters.

        Args:
            project_id: Identifier of the project
            content: URL or local file path to a video file
            content_array: List of URLs or local file paths to video files
            processing_parameters: Video processing configuration
            processing_parameters_array: List of video processing configurations for each asset
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single video asset
            >>> result = kili.assets.create_video_native(
            ...     project_id="my_project",
            ...     content="https://example.com/video.mp4"
            ... )

            >>> # Create video with processing parameters
            >>> result = kili.assets.create_video_native(
            ...     project_id="my_project",
            ...     content="https://example.com/video.mp4",
            ...     processing_parameters={"frames_played_per_second": 25}
            ... )

            >>> # Create multiple video assets
            >>> result = kili.assets.create_video_native(
            ...     project_id="my_project",
            ...     content_array=["https://example.com/video1.mp4", "https://example.com/video2.mp4"],
            ...     processing_parameters_array=[{"frames_played_per_second": 25}, {"frames_played_per_second": 30}]
            ... )
        """
        # Convert singular to plural
        if content is not None:
            content_array = cast(Union[list[str], list[dict]], [content])
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]
        if processing_parameters is not None:
            processing_parameters_array = [processing_parameters]

        # Merge processing parameters into json_metadata
        if processing_parameters_array is not None:
            if json_metadata_array is None:
                json_metadata_array = [{} for _ in processing_parameters_array]
            for i, params in enumerate(processing_parameters_array):
                if i < len(json_metadata_array):
                    json_metadata_array[i][
                        "processingParameters"
                    ] = _prepare_video_processing_parameters(params, use_native_video=True)

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            content_array=content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_video_frame(
        self,
        *,
        project_id: str,
        json_content: Union[List[Union[dict, str]], None],
        processing_parameters: Optional[VideoProcessingParameters] = None,
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_video_frame(
        self,
        *,
        project_id: str,
        json_content_array: List[Union[List[Union[dict, str]], None]],
        processing_parameters_array: Optional[List[VideoProcessingParameters]] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_video_frame(
        self,
        *,
        project_id: str,
        json_content: Optional[Union[List[Union[dict, str]], None]] = None,
        json_content_array: Optional[List[Union[List[Union[dict, str]], None]]] = None,
        processing_parameters: Optional[VideoProcessingParameters] = None,
        processing_parameters_array: Optional[List[VideoProcessingParameters]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create video assets from frame sequences in a project.

        If processing parameters are incomplete, Kili will probe the videos to determine missing parameters.

        Args:
            project_id: Identifier of the project
            json_content: Sequence of frames (list of URLs or paths to images)
            json_content_array: List of frame sequences for each video
            processing_parameters: Video processing configuration
            processing_parameters_array: List of video processing configurations for each asset
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single video from frames
            >>> result = kili.assets.create_video_frame(
            ...     project_id="my_project",
            ...     json_content=["https://example.com/frame1.png", "https://example.com/frame2.png"]
            ... )

            >>> # Create video from frames with processing parameters
            >>> result = kili.assets.create_video_frame(
            ...     project_id="my_project",
            ...     json_content=["https://example.com/frame1.png", "https://example.com/frame2.png"],
            ...     processing_parameters={"frames_played_per_second": 25}
            ... )

            >>> # Create multiple videos from frames
            >>> result = kili.assets.create_video_frame(
            ...     project_id="my_project",
            ...     json_content_array=[
            ...         ["https://example.com/video1/frame1.png", "https://example.com/video1/frame2.png"],
            ...         ["https://example.com/video2/frame1.png", "https://example.com/video2/frame2.png"]
            ...     ]
            ... )
        """
        # Convert singular to plural
        if json_content is not None:
            json_content_array = [json_content]
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]
        if processing_parameters is not None:
            processing_parameters_array = [processing_parameters]

        # Merge processing parameters into json_metadata
        if processing_parameters_array is not None:
            if json_metadata_array is None:
                json_metadata_array = [{} for _ in processing_parameters_array]
            for i, params in enumerate(processing_parameters_array):
                if i < len(json_metadata_array):
                    json_metadata_array[i][
                        "processingParameters"
                    ] = _prepare_video_processing_parameters(params, use_native_video=False)

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            json_content_array=json_content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_geospatial(
        self,
        *,
        project_id: str,
        layer_array: Optional[List[GeospatialLayerParam]] = None,
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_geospatial(
        self,
        *,
        project_id: str,
        layer_arrays: Optional[List[List[GeospatialLayerParam]]] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_geospatial(
        self,
        *,
        project_id: str,
        layer_array: Optional[List[GeospatialLayerParam]] = None,
        layer_arrays: Optional[List[List[GeospatialLayerParam]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create single/multi-layer geospatial imagery assets in a project.

        Args:
            project_id: Identifier of the project
            layer_array: List of layer paths for a single geospatial asset
            layer_arrays: List of multi-layer content for each geospatial asset
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single geospatial asset
            >>> result = kili.assets.create_geospatial(
            ...     project_id="my_project",
            ...     layer_array=[
            ...         {"path": "/path/to/layer1.tif"},
            ...         {"path": "/path/to/layer2.tif"}
            ...     ]
            ... )

            >>> # Create multiple geospatial assets
            >>> result = kili.assets.create_geospatial(
            ...     project_id="my_project",
            ...     layer_arrays=[
            ...         [{"path": "/path/to/asset1/layer1.tif"}, {"path": "/path/to/asset1/layer2.tif"}],
            ...         [{"path": "/path/to/asset2/layer1.tif"}]
            ...     ]
            ... )
        """
        # Convert singular to plural
        if layer_array is not None:
            layer_arrays = [layer_array]
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        multi_layer_content_array = []
        json_content_array = []

        if layer_arrays is None:
            raise ValueError(
                "One of layer_array or layer_arrays parameter need to be filled"
                " to create a geospatial asset"
            )

        # Split layers into 2 arrays, depending on path parameter
        for layers in layer_arrays:
            multi_layer_contents = []
            json_contents = []
            for layer in layers:
                if is_url(layer.get("path")):
                    # is web url, check mandatory params
                    web_layer = convert_to_web_layer(layer)
                    json_contents.append(web_layer)
                else:
                    local_layer = convert_to_local_layer(layer)
                    multi_layer_contents.append(local_layer)
            multi_layer_content_array.append(multi_layer_contents)
            json_content_array.append(json_contents)

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            multi_layer_content_array=multi_layer_content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            json_content_array=json_content_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_pdf(
        self,
        *,
        project_id: str,
        content: Union[str, dict],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_pdf(
        self,
        *,
        project_id: str,
        content_array: Union[List[str], List[dict]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_pdf(
        self,
        *,
        project_id: str,
        content: Optional[Union[str, dict]] = None,
        content_array: Optional[Union[List[str], List[dict]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create PDF assets in a project.

        Args:
            project_id: Identifier of the project
            content: URL or local file path to a PDF
            content_array: List of URLs or local file paths to PDFs
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single PDF asset
            >>> result = kili.assets.create_pdf(
            ...     project_id="my_project",
            ...     content="https://example.com/document.pdf"
            ... )

            >>> # Create multiple PDF assets
            >>> result = kili.assets.create_pdf(
            ...     project_id="my_project",
            ...     content_array=["https://example.com/doc1.pdf", "https://example.com/doc2.pdf"]
            ... )

            >>> # Create PDF with metadata
            >>> result = kili.assets.create_pdf(
            ...     project_id="my_project",
            ...     content="https://example.com/document.pdf",
            ...     json_metadata={"title": "Contract Document"}
            ... )
        """
        # Convert singular to plural
        if content is not None:
            content_array = cast(Union[list[str], list[dict]], [content])
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            content_array=content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_text(
        self,
        *,
        project_id: str,
        content: Union[str, dict],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_text(
        self,
        *,
        project_id: str,
        content_array: Union[List[str], List[dict]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_text(
        self,
        *,
        project_id: str,
        content: Optional[Union[str, dict]] = None,
        content_array: Optional[Union[List[str], List[dict]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create plain text assets in a project.

        Args:
            project_id: Identifier of the project
            content: Raw text content or URL to text asset
            content_array: List of raw text contents or URLs to text assets
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single text asset
            >>> result = kili.assets.create_text(
            ...     project_id="my_project",
            ...     content="This is a sample text for annotation."
            ... )

            >>> # Create multiple text assets
            >>> result = kili.assets.create_text(
            ...     project_id="my_project",
            ...     content_array=["First text sample", "Second text sample"]
            ... )

            >>> # Create text asset with metadata
            >>> result = kili.assets.create_text(
            ...     project_id="my_project",
            ...     content="Sample text",
            ...     json_metadata={"source": "user_feedback"}
            ... )
        """
        # Convert singular to plural
        if content is not None:
            content_array = cast(Union[list[str], list[dict]], [content])
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            content_array=content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_rich_text(
        self,
        *,
        project_id: str,
        json_content: Union[List[Union[dict, str]], None],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_rich_text(
        self,
        *,
        project_id: str,
        json_content_array: List[Union[List[Union[dict, str]], None]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_rich_text(
        self,
        *,
        project_id: str,
        json_content: Optional[Union[List[Union[dict, str]], None]] = None,
        json_content_array: Optional[List[Union[List[Union[dict, str]], None]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create rich-text formatted text assets in a project.

        Rich-text assets use a structured JSON format to represent formatted text content.
        See the Kili documentation for the rich-text format specification.

        Args:
            project_id: Identifier of the project
            json_content: Rich-text formatted content (JSON structure)
            json_content_array: List of rich-text formatted contents
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single rich-text asset
            >>> result = kili.assets.create_rich_text(
            ...     project_id="my_project",
            ...     json_content=[{"text": "Hello ", "style": "normal"}, {"text": "world", "style": "bold"}]
            ... )

            >>> # Create multiple rich-text assets
            >>> result = kili.assets.create_rich_text(
            ...     project_id="my_project",
            ...     json_content_array=[
            ...         [{"text": "First document", "style": "normal"}],
            ...         [{"text": "Second document", "style": "italic"}]
            ...     ]
            ... )

        !!! info "Rich-text format"
            For detailed information on the rich-text format, see the
            [Kili documentation on importing text assets](
                https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/import_text_assets/
            ).
        """
        # Convert singular to plural
        if json_content is not None:
            json_content_array = [json_content]
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        # Call the legacy method directly through the client
        return self._client.append_many_to_dataset(
            project_id=project_id,
            json_content_array=json_content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def create_audio(
        self,
        *,
        project_id: str,
        content: Union[str, dict],
        external_id: Optional[str] = None,
        json_metadata: Optional[dict] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @overload
    def create_audio(
        self,
        *,
        project_id: str,
        content_array: Union[List[str], List[dict]],
        external_id_array: Optional[List[str]] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        ...

    @typechecked
    def create_audio(
        self,
        *,
        project_id: str,
        content: Optional[Union[str, dict]] = None,
        content_array: Optional[Union[List[str], List[dict]]] = None,
        external_id: Optional[str] = None,
        external_id_array: Optional[List[str]] = None,
        json_metadata: Optional[dict] = None,
        json_metadata_array: Optional[List[dict]] = None,
        disable_tqdm: Optional[bool] = None,
        wait_until_availability: bool = True,
        **kwargs,
    ) -> dict[Literal["id", "asset_ids"], Union[str, List[str]]]:
        """Create audio assets in a project.

        Args:
            project_id: Identifier of the project
            content: URL or local file path to an audio file
            content_array: List of URLs or local file paths to audio files
            external_id: External id to identify the asset
            external_id_array: List of external ids given to identify the assets
            json_metadata: The metadata given to the asset
            json_metadata_array: The metadata given to each asset
            disable_tqdm: If True, the progress bar will be disabled
            wait_until_availability: If True, waits until assets are fully processed
            **kwargs: Additional arguments (e.g., is_honeypot)

        Returns:
            A dictionary with project id and list of created asset ids

        Examples:
            >>> # Create single audio asset
            >>> result = kili.assets.create_audio(
            ...     project_id="my_project",
            ...     content="https://example.com/audio.mp3"
            ... )

            >>> # Create multiple audio assets
            >>> result = kili.assets.create_audio(
            ...     project_id="my_project",
            ...     content_array=["https://example.com/audio1.mp3", "https://example.com/audio2.wav"]
            ... )

            >>> # Create audio with metadata
            >>> result = kili.assets.create_audio(
            ...     project_id="my_project",
            ...     content="https://example.com/audio.mp3",
            ...     json_metadata={"speaker": "John Doe"}
            ... )
        """
        if content is not None:
            content_array = cast(Union[list[str], list[dict]], [content])
        if external_id is not None:
            external_id_array = [external_id]
        if json_metadata is not None:
            json_metadata_array = [json_metadata]

        return self._client.append_many_to_dataset(
            project_id=project_id,
            content_array=content_array,
            external_id_array=external_id_array,
            json_metadata_array=json_metadata_array,
            disable_tqdm=disable_tqdm,
            wait_until_availability=wait_until_availability,
            **kwargs,
        )

    @overload
    def delete(
        self,
        *,
        asset_id: str,
        project_id: str = "",
    ) -> Optional[dict[Literal["id"], str]]:
        ...

    @overload
    def delete(
        self,
        *,
        asset_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[Literal["id"], str]]:
        ...

    @overload
    def delete(
        self,
        *,
        external_id: str,
        project_id: str = "",
    ) -> Optional[dict[Literal["id"], str]]:
        ...

    @overload
    def delete(
        self,
        *,
        external_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[Literal["id"], str]]:
        ...

    @typechecked
    def delete(
        self,
        *,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
    ) -> Optional[dict[Literal["id"], str]]:
        """Delete assets from a project.

        Args:
            asset_id: The asset internal ID to delete.
            asset_ids: The list of asset internal IDs to delete.
            external_id: The asset external ID to delete.
            external_ids: The list of asset external IDs to delete.
            project_id: The project ID. Only required if `external_id(s)` argument is provided.

        Returns:
            A dict object with the project `id`.

        Examples:
            >>> # Delete single asset by internal ID
            >>> result = kili.assets.delete(asset_id="ckg22d81r0jrg0885unmuswj8")

            >>> # Delete multiple assets by internal IDs
            >>> result = kili.assets.delete(
            ...     asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"]
            ... )

            >>> # Delete assets by external IDs
            >>> result = kili.assets.delete(
            ...     external_ids=["asset1", "asset2"],
            ...     project_id="my_project"
            ... )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]

        # Call the legacy method directly through the client
        return self._client.delete_many_from_dataset(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
        )

    @overload
    def update_processing_parameter(
        self,
        *,
        asset_id: str,
        processing_parameter: Union[dict, str],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_processing_parameter(
        self,
        *,
        asset_ids: List[str],
        processing_parameters: List[Union[dict, str]],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_processing_parameter(
        self,
        *,
        external_id: str,
        processing_parameter: Union[dict, str],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_processing_parameter(
        self,
        *,
        external_ids: List[str],
        processing_parameters: List[Union[dict, str]],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @typechecked
    def update_processing_parameter(
        self,
        *,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        processing_parameter: Optional[Union[dict, str]] = None,
        processing_parameters: Optional[List[Union[dict, str]]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        """Update processing_parameter of one or more assets.

        Args:
            asset_id: The internal asset ID to modify.
            asset_ids: The internal asset IDs to modify.
            processing_parameter: Video processing parameter for the asset.
            processing_parameters: Video processing parameters for the assets.
            external_id: The external asset ID to modify (if `asset_id` is not already provided).
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            project_id: The project ID.
            **kwargs: Additional update parameters.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> result = kili.assets.update_processing_parameter(
            ...     asset_id="ckg22d81r0jrg0885unmuswj8",
            ...     processing_parameter={
            ...         "frames_played_per_second": 25,
            ...         "shouldKeepNativeFrameRate": True,
            ...         "shouldUseNativeVideo": True,
            ...         "codec": "h264",
            ...         "delayDueToMinPts": 0,
            ...         "numberOfFrames": 450,
            ...         "startTime": 0
            ...     }
            ... )

            >>> # Multiple assets
            >>> result = kili.assets.update_processing_parameter(
            ...     asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
            ...     processing_parameters=[{
            ...         "frames_played_per_second": 25,
            ...         "shouldKeepNativeFrameRate": True,
            ...     }, {
            ...         "frames_played_per_second": 30,
            ...         "shouldKeepNativeFrameRate": False,
            ...     }]
            ... )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]
        if processing_parameter is not None:
            processing_parameters = [processing_parameter]

        json_metadatas = []
        for p in processing_parameters if processing_parameters is not None else []:
            json_metadatas.append({"processingParameters": p})

        # Call the legacy method directly through the client
        return self._client.update_properties_in_assets(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
            json_metadatas=json_metadatas,
            **kwargs,
        )

    @overload
    def update_external_id(
        self,
        *,
        new_external_id: str,
        asset_id: str,
        project_id: str = "",
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_external_id(
        self,
        *,
        new_external_ids: List[str],
        asset_ids: List[str],
        project_id: str = "",
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_external_id(
        self,
        *,
        new_external_id: str,
        external_id: str,
        project_id: str = "",
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_external_id(
        self,
        *,
        new_external_ids: List[str],
        external_ids: List[str],
        project_id: str = "",
    ) -> List[dict[Literal["id"], str]]:
        ...

    @typechecked
    def update_external_id(
        self,
        *,
        new_external_id: Optional[str] = None,
        new_external_ids: Optional[List[str]] = None,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
    ) -> List[dict[Literal["id"], str]]:
        """Update the external ID of one or more assets.

        Args:
            new_external_id: The new external ID of the asset.
            new_external_ids: The new external IDs of the assets.
            asset_id: The asset ID to modify.
            asset_ids: The asset IDs to modify.
            external_id: The external asset ID to modify (if `asset_id` is not already provided).
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            project_id: The project ID. Only required if `external_id(s)` argument is provided.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> kili.assets.update_external_id(
                    new_external_id="new_asset1",
                    asset_id="ckg22d81r0jrg0885unmuswj8",
                )

            >>> # Multiple assets
            >>> kili.assets.update_external_id(
                    new_external_ids=["asset1", "asset2"],
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                )
        """
        # Convert singular to plural
        if new_external_id is not None:
            new_external_ids = [new_external_id]
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]

        assert new_external_ids is not None, "new_external_ids must be provided"

        return self._client.change_asset_external_ids(
            new_external_ids=new_external_ids,
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
        )

    @overload
    def add_metadata(
        self,
        *,
        json_metadata: dict[str, Union[str, int, float]],
        project_id: str,
        asset_id: str,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def add_metadata(
        self,
        *,
        json_metadata: List[dict[str, Union[str, int, float]]],
        project_id: str,
        asset_ids: List[str],
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def add_metadata(
        self,
        *,
        json_metadata: dict[str, Union[str, int, float]],
        project_id: str,
        external_id: str,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def add_metadata(
        self,
        *,
        json_metadata: List[dict[str, Union[str, int, float]]],
        project_id: str,
        external_ids: List[str],
    ) -> List[dict[Literal["id"], str]]:
        ...

    @typechecked
    def add_metadata(
        self,
        *,
        json_metadata: Union[
            dict[str, Union[str, int, float]], List[dict[str, Union[str, int, float]]]
        ],
        project_id: str,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
    ) -> List[dict[Literal["id"], str]]:
        """Add metadata to assets without overriding existing metadata.

        Args:
            json_metadata: Metadata dictionary to add to asset, or list of metadata dictionaries to add to each asset.
                Each dictionary contains key/value pairs to be added to the asset's metadata.
            project_id: The project ID.
            asset_id: The asset ID to modify.
            asset_ids: The asset IDs to modify.
            external_id: The external asset ID to modify (if `asset_id` is not already provided).
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> kili.assets.add_metadata(
                    json_metadata={"key1": "value1", "key2": "value2"},
                    project_id="cm92to3cx012u7l0w6kij9qvx",
                    asset_id="ckg22d81r0jrg0885unmuswj8"
                )

            >>> # Multiple assets
            >>> kili.assets.add_metadata(
                    json_metadata=[
                        {"key1": "value1", "key2": "value2"},
                        {"key3": "value3"}
                    ],
                    project_id="cm92to3cx012u7l0w6kij9qvx",
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"]
                )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]
        if isinstance(json_metadata, dict):
            json_metadata = [json_metadata]

        return self._client.add_metadata(
            json_metadata=json_metadata,
            project_id=project_id,
            asset_ids=asset_ids,
            external_ids=external_ids,
        )

    @overload
    def set_metadata(
        self,
        *,
        json_metadata: dict[str, Union[str, int, float]],
        project_id: str,
        asset_id: str,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def set_metadata(
        self,
        *,
        json_metadata: List[dict[str, Union[str, int, float]]],
        project_id: str,
        asset_ids: List[str],
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def set_metadata(
        self,
        *,
        json_metadata: dict[str, Union[str, int, float]],
        project_id: str,
        external_id: str,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def set_metadata(
        self,
        *,
        json_metadata: List[dict[str, Union[str, int, float]]],
        project_id: str,
        external_ids: List[str],
    ) -> List[dict[Literal["id"], str]]:
        ...

    @typechecked
    def set_metadata(
        self,
        *,
        json_metadata: Union[
            dict[str, Union[str, int, float]], List[dict[str, Union[str, int, float]]]
        ],
        project_id: str,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
    ) -> List[dict[Literal["id"], str]]:
        """Set metadata on assets, replacing any existing metadata.

        Args:
            json_metadata: Metadata dictionary to set on asset, or list of metadata dictionaries to set on each asset.
                Each dictionary contains key/value pairs to be set as the asset's metadata.
            project_id: The project ID.
            asset_id: The asset ID to modify.
            asset_ids: The asset IDs to modify (if `external_ids` is not already provided).
            external_id: The external asset ID to modify (if `asset_id` is not already provided).
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> kili.assets.set_metadata(
                    json_metadata={"key1": "value1", "key2": "value2"},
                    project_id="cm92to3cx012u7l0w6kij9qvx",
                    asset_id="ckg22d81r0jrg0885unmuswj8"
                )

            >>> # Multiple assets
            >>> kili.assets.set_metadata(
                    json_metadata=[
                        {"key1": "value1", "key2": "value2"},
                        {"key3": "value3"}
                    ],
                    project_id="cm92to3cx012u7l0w6kij9qvx",
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"]
                )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]
        if isinstance(json_metadata, dict):
            json_metadata = [json_metadata]

        return self._client.set_metadata(
            json_metadata=json_metadata,
            project_id=project_id,
            asset_ids=asset_ids,
            external_ids=external_ids,
        )

    @typechecked
    def skip(
        self,
        *,
        asset_id: str,
        project_id: str,
        reason: str,
    ) -> str:
        """Skip an asset.

        Args:
            asset_id: ID of the asset you want to skip.
            project_id: The project ID.
            reason: The reason why you skip an asset.

        Returns:
            The asset ID of the asset skipped. An error message if mutation failed.

        Examples:
            >>> kili.assets.skip(
                project_id="ckg22d81r0jrg0885unmuswj8",
                asset_id="ckg22d81s0jrh0885pdxfd03n",
                reason="Test"
                )
        """
        return self._client.skip_or_unskip(
            action="skip",
            asset_id=asset_id,
            project_id=project_id,
            reason=reason,
        )

    @typechecked
    def unskip(
        self,
        *,
        asset_id: str,
        project_id: str,
    ) -> str:
        """Unskip an asset.

        Args:
            asset_id: ID of the asset you want to unskip.
            project_id: The project ID.

        Returns:
            The asset ID of the asset unskipped. An error message if mutation failed.

        Examples:
            >>> kili.assets.unskip(
                asset_id="ckg22d81s0jrh0885pdxfd03n"
                project_id="ckg22d81r0jrg0885unmuswj8"
            )
        """
        return self._client.skip_or_unskip(
            action="unskip", asset_id=asset_id, project_id=project_id
        )

    @overload
    def invalidate(
        self,
        *,
        external_id: str,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def invalidate(
        self,
        *,
        external_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def invalidate(
        self,
        *,
        asset_id: str,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def invalidate(
        self,
        *,
        asset_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @typechecked
    def invalidate(
        self,
        *,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        """Send assets back to queue (invalidate current step).

        This method sends assets back to the queue, effectively invalidating their
        current workflow step status.

        Args:
            asset_id: Internal ID of asset to send back to queue.
            asset_ids: List of internal IDs of assets to send back to queue.
            external_id: External ID of asset to send back to queue.
            external_ids: List of external IDs of assets to send back to queue.
            project_id: The project ID. Only required if `external_id(s)` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to queue.
            An error message if mutation failed.

        Examples:
            >>> # Single asset
            >>> kili.assets.invalidate(asset_id="ckg22d81r0jrg0885unmuswj8")

            >>> # Multiple assets
            >>> kili.assets.invalidate(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"]
                )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]

        return self._client.send_back_to_queue(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
        )

    @overload
    def move_to_next_step(
        self,
        *,
        asset_id: str,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def move_to_next_step(
        self,
        *,
        asset_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def move_to_next_step(
        self,
        *,
        external_id: str,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @overload
    def move_to_next_step(
        self,
        *,
        external_ids: List[str],
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        ...

    @typechecked
    def move_to_next_step(
        self,
        *,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
    ) -> Optional[dict[str, Any]]:
        """Move assets to the next workflow step (typically review).

        This method moves assets to the next step in the workflow, typically
        adding them to review.

        Args:
            asset_id: The asset internal ID to add to review.
            asset_ids: The asset internal IDs to add to review.
            external_id: The asset external ID to add to review.
            external_ids: The asset external IDs to add to review.
            project_id: The project ID. Only required if `external_id(s)` argument is provided.

        Returns:
            A dict object with the project `id` and the `asset_ids` of assets moved to review.
            `None` if no assets have changed status (already had `TO_REVIEW` status for example).
            An error message if mutation failed.

        Examples:
            >>> # Single asset
            >>> kili.assets.move_to_next_step(asset_id="ckg22d81r0jrg0885unmuswj8")

            >>> # Multiple assets
            >>> kili.assets.move_to_next_step(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"]
                )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]

        return self._client.add_to_review(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
        )

    @overload
    def assign(
        self,
        *,
        to_be_labeled_by: List[str],
        asset_id: str,
        project_id: str = "",
    ) -> List[dict[str, Any]]:
        ...

    @overload
    def assign(
        self,
        *,
        to_be_labeled_by_array: List[List[str]],
        asset_ids: List[str],
        project_id: str = "",
    ) -> List[dict[str, Any]]:
        ...

    @overload
    def assign(
        self,
        *,
        to_be_labeled_by: List[str],
        external_id: str,
        project_id: str = "",
    ) -> List[dict[str, Any]]:
        ...

    @overload
    def assign(
        self,
        *,
        to_be_labeled_by_array: List[List[str]],
        external_ids: List[str],
        project_id: str = "",
    ) -> List[dict[str, Any]]:
        ...

    @typechecked
    def assign(
        self,
        *,
        to_be_labeled_by: Optional[List[str]] = None,
        to_be_labeled_by_array: Optional[List[List[str]]] = None,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
    ) -> List[dict[str, Any]]:
        """Assign a list of assets to a list of labelers.

        Args:
            to_be_labeled_by: List of labeler user IDs to assign to a single asset.
            to_be_labeled_by_array: Array of lists of labelers to assign per asset (list of userIds).
            asset_id: The internal asset ID to assign.
            asset_ids: The internal asset IDs to assign.
            external_id: The external asset ID to assign (if `asset_id` is not already provided).
            external_ids: The external asset IDs to assign (if `asset_ids` is not already provided).
            project_id: The project ID. Only required if `external_id(s)` argument is provided.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> kili.assets.assign(
                    asset_id="ckg22d81r0jrg0885unmuswj8",
                    to_be_labeled_by=['cm3yja6kv0i698697gcil9rtk','cm3yja6kv0i000000gcil9rtk']
                )

            >>> # Multiple assets
            >>> kili.assets.assign(
                    asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
                    to_be_labeled_by_array=[['cm3yja6kv0i698697gcil9rtk','cm3yja6kv0i000000gcil9rtk'],
                                            ['cm3yja6kv0i698697gcil9rtk']]
                )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]
        if to_be_labeled_by is not None:
            to_be_labeled_by_array = [to_be_labeled_by]

        assert to_be_labeled_by_array is not None, "to_be_labeled_by_array must be provided"

        return self._client.assign_assets_to_labelers(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
            to_be_labeled_by_array=to_be_labeled_by_array,
        )

    @overload
    def update_priority(
        self,
        *,
        asset_id: str,
        priority: int,
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_priority(
        self,
        *,
        asset_ids: List[str],
        priorities: List[int],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_priority(
        self,
        *,
        external_id: str,
        priority: int,
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @overload
    def update_priority(
        self,
        *,
        external_ids: List[str],
        priorities: List[int],
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        ...

    @typechecked
    def update_priority(
        self,
        *,
        asset_id: Optional[str] = None,
        asset_ids: Optional[List[str]] = None,
        priority: Optional[int] = None,
        priorities: Optional[List[int]] = None,
        external_id: Optional[str] = None,
        external_ids: Optional[List[str]] = None,
        project_id: str = "",
        **kwargs,
    ) -> List[dict[Literal["id"], str]]:
        """Update the priority of one or more assets.

        Args:
            asset_id: The internal asset ID to modify.
            asset_ids: The internal asset IDs to modify.
            priority: Change the priority of the asset.
            priorities: Change the priority of the assets.
            external_id: The external asset ID to modify (if `asset_id` is not already provided).
            external_ids: The external asset IDs to modify (if `asset_ids` is not already provided).
            project_id: The project ID. Only required if `external_id(s)` argument is provided.
            **kwargs: Additional update parameters.

        Returns:
            A list of dictionaries with the asset ids.

        Examples:
            >>> # Single asset
            >>> result = kili.assets.update_priority(
            ...     asset_id="ckg22d81r0jrg0885unmuswj8",
            ...     priority=1,
            ... )

            >>> # Multiple assets
            >>> result = kili.assets.update_priority(
            ...     asset_ids=["ckg22d81r0jrg0885unmuswj8", "ckg22d81s0jrh0885pdxfd03n"],
            ...     priorities=[1, 2],
            ... )
        """
        # Convert singular to plural
        if asset_id is not None:
            asset_ids = [asset_id]
        if external_id is not None:
            external_ids = [external_id]
        if priority is not None:
            priorities = [priority]

        # Call the legacy method directly through the client
        return self._client.update_properties_in_assets(
            asset_ids=asset_ids,
            external_ids=external_ids,
            project_id=project_id,
            priorities=priorities if priorities is not None else [],
            **kwargs,
        )

    @overload
    def update_consensus(
        self,
        *,
        asset_id: str,
        project_id: str,
        is_consensus: bool,
    ) -> bool:
        ...

    @overload
    def update_consensus(
        self,
        *,
        external_id: str,
        project_id: str,
        is_consensus: bool,
    ) -> bool:
        ...

    @typechecked
    def update_consensus(
        self,
        *,
        project_id: str,
        is_consensus: bool,
        asset_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ) -> bool:
        """Activate or deactivate consensus on an asset.

        Args:
            project_id: The project ID.
            is_consensus: Whether to activate (True) or deactivate (False) consensus on the asset.
            asset_id: The internal asset ID to modify. Either asset_id or external_id must be provided.
            external_id: The external ID of the asset to modify. Either asset_id or external_id must be provided.

        Returns:
            The consensus value that was set (True if consensus was activated, False if deactivated).

        Raises:
            ValueError: If neither asset_id nor external_id is provided.

        Examples:
            >>> # Activate consensus on an asset using asset_id
            >>> result = kili.assets.update_consensus(
            ...     project_id="my_project",
            ...     is_consensus=True,
            ...     asset_id="ckg22d81r0jrg0885unmuswj8"
            ... )
            >>> # result is True

            >>> # Activate consensus on an asset using external_id
            >>> result = kili.assets.update_consensus(
            ...     project_id="my_project",
            ...     is_consensus=True,
            ...     external_id="my_asset_001"
            ... )
            >>> # result is True

            >>> # Deactivate consensus on an asset
            >>> result = kili.assets.update_consensus(
            ...     project_id="my_project",
            ...     is_consensus=False,
            ...     asset_id="ckg22d81r0jrg0885unmuswj8"
            ... )
            >>> # result is False
        """
        return self._client.update_asset_consensus(
            project_id=project_id,
            is_consensus=is_consensus,
            asset_id=asset_id,
            external_id=external_id,
        )



"""Class to upload a plugin."""

import ast
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union
from zipfile import ZipFile

from typing_extensions import LiteralString

from kili.adapters.http_client import HttpClient
from kili.core.constants import (
    mime_extensions_for_py_scripts,
    mime_extensions_for_txt_files,
)
from kili.core.graphql.operations.plugin.mutations import (
    GQL_CREATE_PLUGIN,
    GQL_CREATE_PLUGIN_RUNNER,
    GQL_CREATE_WEBHOOK,
    GQL_GENERATE_UPDATE_URL,
    GQL_UPDATE_PLUGIN_RUNNER,
    GQL_UPDATE_WEBHOOK,
)
from kili.core.graphql.operations.plugin.queries import GQL_GET_PLUGIN_RUNNER_STATUS
from kili.core.helpers import get_mime_type
from kili.services.plugins.tools import check_errors_plugin_upload
from kili.utils import bucket
from kili.utils.tempfile import TemporaryDirectory

from .exceptions import PluginCreationError
from .helpers import get_logger

if TYPE_CHECKING:
    from kili.client import Kili

NUMBER_TRIES_RUNNER_STATUS = 20

POSSIBLE_HANDLERS = {
    "on_submit": "onSubmit",
    "on_review": "onReview",
    "on_custom_interface_click": "onCustomInterfaceClick",
    "on_project_updated": "onProjectUpdated",
    "on_send_back_to_queue": "onSendBackToQueue",
}


def check_file_mime_type(
    path: Path, compatible_mime_extensions: list[str], verbose: bool = True
) -> bool:
    """Returns true if the mime type of the file corresponds to one of compatible_mime_extensions."""
    mime_type = get_mime_type(path.as_posix())

    if not compatible_mime_extensions or mime_type is None:
        return False

    correct_mime_type = mime_type in compatible_mime_extensions

    if verbose and not correct_mime_type:
        print(
            f"File mime type for {path} is {mime_type} and does not correspond"
            "to the type of the project. "
            f"File mime type should be one of {compatible_mime_extensions}"
        )
    return correct_mime_type


def check_file_is_py(path: Path, verbose: bool = True) -> bool:
    """Returns true if the mime type of the file corresponds to a python file."""
    return check_file_mime_type(path, mime_extensions_for_py_scripts, verbose)


def check_file_is_txt(path: Path, verbose: bool = True) -> bool:
    """Returns true if the mime type of the file corresponds to a .txt file."""
    return check_file_mime_type(path, mime_extensions_for_txt_files, verbose)


def check_file_contains_handler(path: Path) -> tuple[bool, Optional[list[str]], bool]:
    """Test if the file is a python file with the right content.

    Return true if the file contain PluginHandler Class.
    Return the list of handlers or None if they are present in the file.
    Return true if the file contains on_event method.
    """
    with path.open(encoding="utf-8") as file:
        module = ast.parse(file.read())
    for node in module.body:
        if isinstance(node, ast.ClassDef) and node.name == "PluginHandler":
            handlers = [
                POSSIBLE_HANDLERS[child.name]
                for child in node.body
                if isinstance(child, ast.FunctionDef) and child.name in POSSIBLE_HANDLERS
            ]
            has_on_event = any(
                isinstance(child, ast.FunctionDef) and child.name == "on_event"
                for child in node.body
            )
            return (True, handlers, has_on_event) if handlers else (True, None, has_on_event)
    return False, None, False


class WebhookUploader:
    """Class to create a webhook."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        kili: "Kili",
        webhook_url: str,
        plugin_name: str,
        header: Optional[str],
        verbose: bool,
        handler_types: Optional[list[str]],
        event_matcher: Optional[list[str]],
    ) -> None:
        self.kili = kili
        self.webhook_url = webhook_url
        self.plugin_name = plugin_name or self.webhook_url
        self.header = header
        self.verbose = verbose
        self.handler_types = handler_types
        self.event_matcher = event_matcher

    def create_webhook(self) -> str:
        """Create a webhook receiving Kili events."""
        variables = {
            "eventMatcher": self.event_matcher,
            "handlerTypes": self.handler_types,
            "pluginName": self.plugin_name,
            "webhookUrl": self.webhook_url,
            "header": self.header,
        }

        result = self.kili.graphql_client.execute(GQL_CREATE_WEBHOOK, variables)

        return self.kili.format_result("data", result)

    def update_webhook(self) -> str:
        """Update a webhook receiving Kili events."""
        variables = {
            "eventMatcher": self.event_matcher,
            "handlerTypes": self.handler_types,
            "pluginName": self.plugin_name,
            "webhookUrl": self.webhook_url,
            "header": self.header,
        }

        result = self.kili.graphql_client.execute(GQL_UPDATE_WEBHOOK, variables)

        return self.kili.format_result("data", result)


class PluginUploader:
    """Class to upload a plugin."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        kili: "Kili",
        plugin_path: str,
        plugin_name: Optional[str],
        verbose: bool,
        http_client: HttpClient,
        event_matcher: Optional[list[str]],
    ) -> None:
        self.kili = kili
        self.plugin_path = Path(plugin_path)
        self.http_client = http_client
        self.event_matcher = event_matcher

        if (not self.plugin_path.is_dir()) and (not self.plugin_path.is_file()):
            raise FileNotFoundError(
                f"The provided path '{plugin_path}' is neither a directory nor a file. The absolute"
                f" path is the following: '{self.plugin_path.absolute()}'"
            )

        if plugin_name:
            self.plugin_name = plugin_name
        else:
            self.plugin_name = self.plugin_path.name
        self.verbose = verbose
        self.handler_types = None

    def _retrieve_plugin_src(self) -> list[Path]:
        """Retrieve script from plugin_path and execute it to prevent an upload with indentation errors."""
        if self.plugin_path.is_dir():
            file_path = self.plugin_path / "main.py"
            if not file_path.is_file():
                raise FileNotFoundError(
                    f"No main.py file in the provided folder: {self.plugin_path.absolute()}"
                )
            contains_handler, handler_types, has_on_event = check_file_contains_handler(file_path)
            if not contains_handler:
                raise ValueError("PluginHandler class is not present in your main.py file.")

            if has_on_event and not self.event_matcher:
                raise ValueError("Event matcher is required for plugins with on_event method.")

            if handler_types and self.event_matcher:
                raise ValueError("Cannot have both handler types and event matcher.")

            self.handler_types = handler_types

            return list(self.plugin_path.glob("**/*.py"))

        file_path = self.plugin_path

        if not check_file_is_py(file_path, self.verbose):
            raise ValueError("Wrong file format.")

        contains_handler, handler_types, has_on_event = check_file_contains_handler(file_path)
        if not contains_handler:
            raise ValueError("PluginHandler class is not present in your plugin file.")

        if has_on_event and not self.event_matcher:
            raise ValueError("Event matcher is required for plugins with on_event method.")

        if handler_types and self.event_matcher:
            raise ValueError("Cannot have both handler types and event matcher.")

        self.handler_types = handler_types

        return [file_path]

    def _retrieve_requirements(self) -> Union[Path, None]:
        """Retrieve script from file_path and execute it to prevent an upload with indentation errors."""
        if not self.plugin_path.is_dir():
            return None

        file_path = self.plugin_path / "requirements.txt"
        if not file_path.is_file():
            logger = get_logger()
            logger.info(
                f"No requirements.txt file in the provided folder: {self.plugin_path.absolute()}"
            )
            return None

        if not check_file_is_txt(file_path, self.verbose):
            raise ValueError("Wrong file format.")

        return file_path

    @staticmethod
    def _parse_script(script_path: Path) -> None:
        """Method to detect indentation and class errors in the script."""
        with script_path.open("r", encoding="utf-8") as file:
            source_code = file.read()

        # We execute the source code to prevent the upload of a file with SyntaxError
        compile(source_code, "<string>", "exec")

    @staticmethod
    def _upload_file(zip_path: Path, url: str, http_client: HttpClient) -> None:
        """Upload a file to a signed url and returns the url with the file_id."""
        bucket.upload_data_via_rest(url, zip_path.read_bytes(), "application/zip", http_client)

    def _retrieve_upload_url(self, is_updating_plugin: bool) -> str:
        """Retrieve an upload url from the backend."""
        variables = {"pluginName": self.plugin_name}

        if is_updating_plugin:
            result = self.kili.graphql_client.execute(GQL_GENERATE_UPDATE_URL, variables)
        else:
            result = self.kili.graphql_client.execute(GQL_CREATE_PLUGIN, variables)

        check_errors_plugin_upload(result, self.plugin_path, self.plugin_name)
        return self.kili.format_result("data", result)

    def _create_zip(self, tmp_directory: Path) -> Path:
        """Create a zip file from python file and requirements.txt.

        (if user has defined a path to a requirements.txt)
        """
        file_paths = self._retrieve_plugin_src()

        for path in file_paths:
            self._parse_script(path)

        requirements_path = self._retrieve_requirements()

        zip_path = tmp_directory / "archive.zip"

        with ZipFile(zip_path, "w") as archive:
            if len(file_paths) == 1:
                archive.write(file_paths[0], "main.py")

            else:
                for path in file_paths:
                    archive.write(path, path.relative_to(self.plugin_path))

            if requirements_path:
                archive.write(requirements_path, requirements_path.relative_to(self.plugin_path))

        return zip_path

    def _upload_script(self, is_updating_plugin: bool) -> None:
        """Upload a script to Kili bucket."""
        with TemporaryDirectory() as tmp_directory:
            zip_path = self._create_zip(tmp_directory)

            upload_url = self._retrieve_upload_url(is_updating_plugin)

            self._upload_file(zip_path, upload_url, self.http_client)

    def _create_plugin_runner(self) -> Any:
        """Create plugin's runner."""
        variables = {
            "pluginName": self.plugin_name,
            "handlerTypes": self.handler_types,
            "eventMatcher": self.event_matcher,
        }

        result = self.kili.graphql_client.execute(GQL_CREATE_PLUGIN_RUNNER, variables)
        return self.kili.format_result("data", result)

    def _check_plugin_runner_status(self, update: bool = False) -> LiteralString:
        """Check the status of a plugin's runner until it is active."""
        logger = get_logger()

        action = "updated" if update else "created"

        logger.info(f"Plugin is being {action}... This should take few minutes.")

        n_tries = 0
        status = None

        while n_tries < NUMBER_TRIES_RUNNER_STATUS:
            time.sleep(15)

            status = self.get_plugin_runner_status()

            if status in ("ACTIVE", "FAILED"):
                break

            logger.info(f"Status : {status}...")

            n_tries += 1

        # this represents the case where the plugin is still deploying
        # after 150 tries (37.5 minutes)
        if status == "DEPLOYING" and n_tries == 150:
            raise RuntimeError(
                "We could not check your plugin was deployed in time. Please check again the"
                " status of the plugin after some minutes with the command:"
                f' kili.get_plugin_status("{self.plugin_name}"). If the status is different than'
                " DEPLOYING or ACTIVE, please check your plugin's code and try to overwrite the"
                " plugin with a new version of the code (you can use kili.update_plugin() for"
                " that)."
            )

        if status != "ACTIVE":
            raise PluginCreationError(
                "There was some error during the creation of the plugin. Please check your"
                " plugin's code and try to overwrite the plugin with a new version of the code"
                " (you can use kili.update_plugin() for that). Hint: You can get build errors"
                f' using: kili.get_plugin_build_errors(plugin_name="{self.plugin_name}")'
            )

        message = f"Plugin {action} successfully"
        logger.info(message)

        return message

    def get_plugin_runner_status(self) -> Any:
        """Get the status of a plugin's runner."""
        variables = {"name": self.plugin_name}

        result = self.kili.graphql_client.execute(GQL_GET_PLUGIN_RUNNER_STATUS, variables)

        return self.kili.format_result("data", result)

    def create_plugin(self):
        """Create a plugin in Kili."""
        self._upload_script(False)

        self._create_plugin_runner()

        return self._check_plugin_runner_status()

    def _update_plugin_runner(self) -> Any:
        """Update plugin's runner."""
        variables = {
            "pluginName": self.plugin_name,
            "handlerTypes": self.handler_types,
            "eventMatcher": self.event_matcher,
        }

        result = self.kili.graphql_client.execute(GQL_UPDATE_PLUGIN_RUNNER, variables)
        return self.kili.format_result("data", result)

    def update_plugin(self):
        """Update a plugin in Kili."""
        self._upload_script(True)

        self._update_plugin_runner()

        return self._check_plugin_runner_status(update=True)



"""Base class for all domain namespaces in the Kili Python SDK.

This module provides the foundational DomainNamespace class that implements
memory optimization, weak references, and caching for all
domain-specific namespaces.
"""

import weakref
from typing import TYPE_CHECKING, Optional, TypeVar

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway

if TYPE_CHECKING:
    from kili.client import Kili as KiliLegacy

T = TypeVar("T", bound="DomainNamespace")


class DomainNamespace:
    """Base class for all domain namespaces with performance optimizations.

    This class provides the foundational architecture for domain-based API namespaces
    in the Kili Python SDK, featuring:

    - Memory efficiency through __slots__
    - Weak references to prevent circular references

    All domain namespaces (assets, labels, projects, etc.) should inherit from this class.
    """

    __slots__ = (
        "_client_ref",
        "_gateway",
        "_domain_name",
        "__weakref__",
    )

    def __init__(
        self,
        client: "KiliLegacy",
        gateway: KiliAPIGateway,
        domain_name: Optional[str] = None,
    ) -> None:
        """Initialize the domain namespace.

        Args:
            client: The Kili client instance
            gateway: The KiliAPIGateway instance for API operations
            domain_name: Optional domain name for debugging/logging purposes
        """
        # Use weak reference to prevent circular references between client and namespaces
        self._client_ref: "weakref.ReferenceType[KiliLegacy]" = weakref.ref(client)
        self._gateway = gateway
        self._domain_name = domain_name or self.__class__.__name__.lower()

    @property
    def _client(self) -> "KiliLegacy":
        """Get the Kili client instance.

        Returns:
            The Kili client instance

        Raises:
            ReferenceError: If the client instance has been garbage collected
        """
        client = self._client_ref()
        if client is None:
            raise ReferenceError(
                f"The Kili client instance for {self._domain_name} namespace "
                "has been garbage collected"
            )
        return client

    def __repr__(self) -> str:
        """Return a string representation of the namespace."""
        try:
            client_info = f"client={self._client.__class__.__name__}"
        except ReferenceError:
            client_info = "client=<garbage collected>"

        return f"{self.__class__.__name__}({client_info}, domain='{self._domain_name}')"



"""Code specific to Azure blob storage."""

import re
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

from azure.storage.blob import BlobServiceClient

from kili.domain.project import InputType


def get_blob_paths_azure_data_connection_with_service_credentials(
    data_connection: dict, data_integration: dict, input_type: InputType
) -> tuple[list[str], list[str], list[str]]:
    """Get the blob paths for an Azure data connection using service credentials."""
    if not (data_integration["azureSASToken"] and data_integration["azureConnectionURL"]):
        raise ValueError(
            f"Cannot retrieve blob paths for data connection {data_connection['id']} with data"
            f" integration {data_integration['id']}. Need to provide \"azureSASToken\" and"
            f' "azureConnectionURL" in data integration: {data_integration}'
        )

    return AzureBucket(
        sas_token=data_integration["azureSASToken"],
        connection_url=data_integration["azureConnectionURL"],
    ).get_blob_paths_azure_data_connection_with_service_credentials(
        input_type=input_type,
        selected_folders=data_connection.get("selectedFolders"),
        prefix=data_connection.get("prefix"),
        include=data_connection.get("include"),
        exclude=data_connection.get("exclude"),
    )


class AzureBucket:
    """Class for Azure blob storage buckets."""

    def __init__(self, sas_token: str, connection_url: str) -> None:
        """Initialize the Azure bucket."""
        self.sas_token = sas_token
        self.connection_url = connection_url

        (
            self.storage_account,
            self.container_name,
        ) = self._split_connection_url_into_storage_account_and_container_name(connection_url)
        self.blob_sas_url = f"https://{self.storage_account}.blob.core.windows.net?{self.sas_token}"
        self.client = BlobServiceClient(account_url=self.blob_sas_url)
        self.storage_bucket = self.client.get_container_client(self.container_name)

    @staticmethod
    def _split_connection_url_into_storage_account_and_container_name(
        connection_url: str,
    ) -> tuple[str, str]:
        """Split the connection url into storage account and container name."""
        split_value = ".blob.core.windows.net"
        url_connection = urlparse(connection_url)
        storage_account = url_connection.hostname.split(  # pyright: ignore[reportOptionalMemberAccess]
            split_value
        )[0]
        container_name = url_connection.path.lstrip("/")
        return storage_account, container_name

    def get_blob_paths_as_tree(self) -> dict:
        """Get a tree representation of the Azure bucket.

        Folder structure is represented as a dictionary.
        A folder is represented as a key with a value as a dictionary.
        A file is represented as a key with a value as None.
        """
        filetree = {}

        for filepath in self.storage_bucket.list_blob_names():
            current_node = filetree
            *folders, filename = Path(filepath).parts
            for folder in folders:
                if folder not in current_node:
                    current_node[folder] = {}
                current_node = current_node[folder]
            if filename:
                current_node[filename] = None

        return filetree

    @staticmethod
    def _generate_regex_pattern(pattern: str) -> str:
        pattern = re.escape(pattern)

        if "*" in pattern:
            pattern = pattern.replace(r"\*", ".+")
            pattern = f"^{pattern}$"
        else:
            pattern = f"^{pattern}"

        return pattern

    def get_blob_paths_azure_data_connection_with_service_credentials(
        self,
        input_type: InputType,
        selected_folders: Optional[list[str]] = None,
        prefix: Optional[str] = None,
        include: Optional[list[str]] = None,
        exclude: Optional[list[str]] = None,
    ) -> tuple[list[str], list[str], list[str]]:
        """Get the blob paths for an Azure data connection using service credentials."""
        blob_paths = []
        content_types = []
        warnings = set()
        include = [self._generate_regex_pattern(pattern) for pattern in include] if include else []
        exclude = [self._generate_regex_pattern(pattern) for pattern in exclude] if exclude else []
        for blob in self.storage_bucket.list_blobs(name_starts_with=prefix):
            if not hasattr(blob, "name") or not isinstance(blob.name, str):
                continue

            if (
                prefix is None
                and isinstance(selected_folders, list)
                and not any(
                    blob.name.startswith(selected_folder) for selected_folder in selected_folders
                )
            ):
                continue

            if len(include) > 0 and not any(re.match(pattern, blob.name) for pattern in include):
                continue

            if len(exclude) > 0 and any(re.match(pattern, blob.name) for pattern in exclude):
                continue

            has_content_type_field = (
                hasattr(blob, "content_settings")
                and hasattr(blob.content_settings, "content_type")
                and isinstance(blob.content_settings.content_type, str)
            )
            if not has_content_type_field:
                warnings.add("Objects with missing content-type were ignored")

            else:
                content_type = blob.content_settings.content_type
                assert isinstance(content_type, str)  # Help type checker understand
                if not self._is_content_type_compatible_with_input_type(
                    content_type,
                    input_type,
                ):
                    warnings.add(
                        "Objects with unsupported content-type for this type of project were ignored"
                    )
                else:
                    blob_paths.append(blob.name)
                    content_types.append(content_type)

        return blob_paths, list(warnings), content_types

    @staticmethod
    def _is_content_type_compatible_with_input_type(
        content_type: str, input_type: InputType
    ) -> bool:
        """Check if the content type is compatible with the input type."""
        if input_type == "IMAGE":
            return content_type.startswith("image")

        if input_type == "VIDEO":
            return content_type.startswith("video")

        if input_type == "PDF":
            return content_type.startswith("application/pdf")

        if input_type == "TEXT":
            return any(
                content_type.startswith(text_type)
                for text_type in (
                    "application/json",
                    "text/plain",
                    "text/html",
                    "text/csv",
                    "text/xml",
                )
            )

        raise ValueError(f"Unknown project input type: {input_type}")
